#!/usr/bin/env python
"""Utilities used by multiple functions / across the library
"""
import warnings
import numpy
from scipy.sparse import coo_matrix

#===============================================================================
# Printing
#===============================================================================


class NullWriter(object):
    """File-like object that actually writes nothing, in the spirit of
    :obj:`os.devnull`
    """

    def write(inp):
        pass

    def __repr__(self):
        return "NullWriter()"

    def __str__(self):
        return "NullWriter()"

    def close(self):
        pass

    def flush(self):
        pass


#===============================================================================
# Building tables when state paths are known
#===============================================================================


def build_hmm_tables(
        num_states,
        state_sequences,
        weights=None,
        state_prior_pseudocounts=0,
        transition_pseudocounts=0,
        normalize=True,
        initializer=numpy.zeros,
):
    """Build a set of state prior and transition tables from sequences of known
    states. This in contrast to *training*, in which parameters for transition
    tables are estimated from msequences of observations and unknown states.


    Parameters
    ----------
    num_states : int
        Number of states

    state_sequences : list of list_like
        Sequences of states, represented as integers

    weights : list-like or None, optional
        Weight to apply to each sequence. If `None`, each sequence will be
        weighted equally (Default: None)

    staet_prior_pseudocounts : int or matrix-like
        Pseudocounts to add to count table. If an ``int``, same value will be
        added to every cell in the table. If matrix or array-like, that matrix
        will be added to the count matrix (Default: 0)

    transition_pseudocounts : int or matrix-like
        Pseudocounts to add to count table. If `int`, same value will be added
        to every cell in the table. If matrix or array-like, that matrix will
        be added to the count matrix (Default: 0)

    normalize : bool, optional
        If ``True``, return a row-normalized transition table. If ``False``,
        return a count table (Default: ``True``)

    initializer : callable
        Initializer for transition table. Must accept `(shape, dtype)` as
        parameters. By default, this would be :func:`numpy.zeros`, to create a
        dense matrix. Another good option would be
        :class:`scipy.sparse.dok_matrix` to create a sparse matrix


    Returns
    -------
    :class:`numpy.ndarray`
        Array of state priors

    `initializer`
        Table[i,j] of transition probabilities from state `i` to state `j`
    """
    tmat = initializer((num_states, num_states), dtype=float)
    state_priors = numpy.zeros(num_states, dtype=float)

    if weights is None:
        weights = [1] * len(state_sequences)

    for my_seq, my_weight in zip(state_sequences, weights):
        if not numpy.isnan(my_weight):
            state_priors[my_seq[0]] += my_weight

            for i in range(len(my_seq) - 1):
                from_seq, to_seq = my_seq[i:i + 2]
                tmat[from_seq, to_seq] += my_weight

    state_priors += state_prior_pseudocounts
    tmat += transition_pseudocounts

    if (tmat.sum(1) == 0).any():
        warnings.warn(
            "There are all-zero rows in the transition table! These may yield "
            "nonsensical probabilities! Consider adding pseudocounts",
            UserWarning
        )

    if normalize == True:
        tmat = (1.0 * tmat.T / tmat.sum(1)).T
        state_priors = state_priors.astype(float) / state_priors.sum()

    return state_priors, tmat


#===============================================================================
# Serialization / deserialization of matrices
#===============================================================================


def matrix_to_dict(mat):
    """Convert a matrix or array `mat` to a sparse dictionary. Not necessary
    for serialization

    Parameters
    ----------
    mat : :class:`numpy.ndarray`, or something like it


    Returns
    -------
    dict
        Dictionary with the following properties:

        `shape`
            A tuple of matrix dimensions

        `row`
            A list of row coordinates

        `col`
            A list of column coordinates

        `data`
            A list of values

    See also
    --------
    matrix_from_dict
    """
    coomat = coo_matrix(mat)
    dout = {
        "shape": tuple(coomat.shape),
        "row": list([int(X) for X in coomat.row]),
        "col": list([int(X) for X in coomat.col]),
        "data": list([float(X) for X in coomat.data]),
    }
    return dout


def matrix_from_dict(dtmp, dense=False):
    """Reconstruct a matrix from a dictionary made e.g. by :func:`matrix_to_dict`


    Parameters
    ----------
    dict : dict
        Dictionary with keys `shape`, `row`, `col`, and `data`

    dense : bool, optional
        Whether or not to return a dense matrix


    Returns
    -------
    :class:`scipy.sparse.coo_matrix` if `dense` is `False`, otherwise :class:`numpy.Matrix`


    See also
    --------
    matrix_to_dict
    """
    coomat = coo_matrix((dtmp["data"], (dtmp["row"], dtmp["col"])), shape=dtmp["shape"])
    return coomat.todense() if dense is True else coomat



#!/usr/bin/env vpython
import sys
import unittest

import numpy
import scipy.stats

from numpy.testing import (
    assert_almost_equal,
)
from nose.plugins.attrib import attr
from nose.tools import assert_greater_equal, assert_true
from minihmm.test.common import check_almost_equal

from minihmm.hmm import FirstOrderHMM

from minihmm.estimators import (
    UnivariateGaussianEmissionEstimator,
    DiscreteEmissionEstimator,
    DiscreteStatePriorEstimator,
    DiscreteTransitionEstimator,
)
from minihmm.training import (train_baum_welch, neg_exp_noise_gen, DefaultLoggerFactory)
from minihmm.factors import (ArrayFactor, MatrixFactor, ScipyDistributionFactor)


class ScreenWriter:
    @staticmethod
    def write(inp):
        sys.stdout.write(inp + "\n")


class _BaseExample:

    @classmethod
    def setUpClass(cls):
        cls.do_subclass_setup()
        print("Setting up tests for %s" % cls.test_name)
        cls.generating_hmm = FirstOrderHMM(
            state_priors=ArrayFactor(cls.arrays["state_priors"]),
            trans_probs=MatrixFactor(cls.arrays["transition_probs"]),
            emission_probs=cls.get_emission_probs()
        )

        cls.naive_hmm = FirstOrderHMM(
            state_priors=ArrayFactor(cls.arrays["naive_state_priors"]),
            trans_probs=MatrixFactor(cls.arrays["naive_transition_probs"]),
            emission_probs=cls.get_naive_emission_probs()
        )

        # generate 100 training examples
        cls.training_examples = [
            cls.generating_hmm.generate(cls.example_len)[1] for _ in range(cls.num_examples)
        ]

        # retrain
        cls.training_results = train_baum_welch(
            cls.naive_hmm,
            cls.training_examples,
            state_prior_estimator=cls.state_prior_estimator,
            transition_estimator=cls.transition_estimator,
            emission_estimator=cls.emission_estimator,
            miniter=80,
            maxiter=1000,
            processes=1,
            logfunc=DefaultLoggerFactory(ScreenWriter(), cls.naive_hmm, maxcols=5),
        )

    @classmethod
    def do_subclass_setup(cls):
        # must define the following in subclasses
        cls.test_name = None
        cls.arrays = {}
        cls.example_len = 200
        cls.num_examples = 100
        cls.state_prior_estimator = None
        cls.transition_estimator = None
        cls.emission_estimator = None

    def test_state_priors_trained(self):
        found = self.training_results["best_model"].state_priors.data
        expected = self.generating_hmm.state_priors.data
        yield check_almost_equal, found, expected, {"decimal": 1}

    def test_transition_probs_trained(self):
        found = self.training_results["best_model"].trans_probs.data
        expected = self.generating_hmm.trans_probs.data
        yield check_almost_equal, found, expected, {"decimal": 2}

    def test_likelihoods_increased_each_round(self):
        delta = numpy.convolve([1, -1], self.training_results["weight_logprobs"], mode="valid")
        assert_greater_equal((delta >= 0).sum() / len(delta), 0.9)

    # override this method in subclass
    def test_emission_probs_trained(self):
        assert False


class TestCasino(_BaseExample):
    """Test re-training on a two-state HMM with discrete emissions"""

    @classmethod
    def get_emission_probs(cls):
        return [ArrayFactor(X) for X in cls.arrays["emission_probs"]]

    @classmethod
    def get_naive_emission_probs(cls):
        return [ArrayFactor(X) for X in cls.arrays["naive_emission_probs"]]

    @classmethod
    def do_subclass_setup(cls):
        cls.test_name = "DiscreteCasino"
        cls.example_len = 200
        cls.num_examples = 150
        cls.state_prior_estimator = DiscreteStatePriorEstimator()
        cls.transition_estimator = DiscreteTransitionEstimator()
        cls.emission_estimator = DiscreteEmissionEstimator(6)

        cls.arrays = {
            "state_priors": numpy.array([0.99, 0.01]),
            "naive_state_priors": numpy.array([0.6, 0.4]),
            "transition_probs": numpy.array([[0.95, 0.05], [0.1, 0.9]]),
            "naive_transition_probs": numpy.array([[0.6, 0.4], [0.5, 0.5]]),
            "emission_probs":[
                1.0 * numpy.ones(6) / 6.0,
                numpy.array([0.1, 0.1, 0.1, 0.1, 0.1, 0.5])
            ],
            "naive_emission_probs": [
                1.0 * numpy.ones(6) / 6.0,
                1.0 * numpy.ones(6) / 6.0,
            ],
        }

    def test_emission_probs_trained(self):
        for found, expected in zip(
               self.training_results["best_model"].emission_probs,
               self.generating_hmm.emission_probs
            ):
            #sse = (found.data - expected.data)**2
            #assert_true((sse < 1e-4).all())
            yield check_almost_equal, found.data, expected.data, {"decimal": 2}



@attr("slow")
class TestGaussian(_BaseExample):
    """Test re-training on a three-state HMM with continuous emissions"""

    @classmethod
    def get_emission_probs(cls):
        return [
            ScipyDistributionFactor(scipy.stats.norm, loc=0, scale=0.5),
            ScipyDistributionFactor(scipy.stats.norm, loc=5, scale=3),
            ScipyDistributionFactor(scipy.stats.norm, loc=-2, scale=1),
        ]

    @classmethod
    def get_naive_emission_probs(cls):
        return [
            ScipyDistributionFactor(scipy.stats.norm, loc=0, scale=1),
            ScipyDistributionFactor(scipy.stats.norm, loc=0, scale=1),
            ScipyDistributionFactor(scipy.stats.norm, loc=0, scale=1),
        ]

    @classmethod
    def do_subclass_setup(cls):
        cls.test_name = "ContinuousGaussian"
        cls.example_len = 500
        cls.num_examples = 150
        cls.state_prior_estimator = DiscreteStatePriorEstimator()
        cls.transition_estimator = DiscreteTransitionEstimator()
        cls.emission_estimator = UnivariateGaussianEmissionEstimator()

        cls.arrays = {
            "state_priors": numpy.array([0.8, 0.05, 0.15]),
            "naive_state_priors": numpy.array([0.6, 0.2, 0.2]),
            "transition_probs": numpy.array([
                [0.9, 0.05, 0.05],
                [0.1, 0.6, 0.3],
                [0.2, 0.1, 0.7],
            ]),
            "naive_transition_probs": numpy.array([
                [0.40, 0.30, 0.30],
                [0.30, 0.36, 0.33],
                [0.33, 0.30, 0.36],
            ]),
        }

    def test_emission_probs_trained(self):
        for found, expected in zip(
            self.training_results["best_model"].emission_probs,
            self.generating_hmm.emission_probs
        ):
            for k in expected.dist_kwargs.keys():
                yield (
                    check_almost_equal,
                    found.dist_kwargs[k],
                    expected.dist_kwargs[k],
                    { "decimal": 2 },
                )





#!/usr/bin/env python
"""A parallelized, extensible implementation of the Baum-Welch training algorithm
built atop plugin estimator classes, which determine how parameters for the HMM
are re-estimated from observation data during training (i.e. the E and M steps
of Expectation maximization).

Also includes helper functions to add decaying noise during training
"""
from minihmm.hmm import FirstOrderHMM
from minihmm.factors import (
    ArrayFactor,
    MatrixFactor,
    FunctionFactor,
    LogFunctionFactor,
    ScipyDistributionFactor,
)
from minihmm.estimators import (
    DiscreteStatePriorEstimator,
    DiscreteTransitionEstimator,
    DiscreteEmissionEstimator,
)
from minihmm.util import NullWriter
import numpy
import datetime
import multiprocessing
import functools
import itertools

#===============================================================================
# INDEX: helper functions
#===============================================================================


def neg_exp_noise_gen(a=1.0, b=1.0, offset=0):
    """Generate exponentially decaying constants following `y = ae**-bx`

    Parameters
    ----------
    a : float, optional
        Initial (maximum) value (Default: 1.0)

    b : float, optional
        Decay constant (Default: 1.0)

    offset : int, optional
        Starting offset in time (in case of training restart)

    Yields
    ------
    float
        Next noise amount
    """
    offset -= 1
    while True:
        offset += 1
        yield a * numpy.exp(-offset * b)


def linear_noise_gen(m=-0.05, b=1, offset=0):
    """Generate linearly decaying noise following `y = max(m*x + b, 0)`

    Parameters
    ----------
    m : float, optional
        Slope of line (default: -0.05)

    b : float, optional
        Y-intercept of line (Default: 1)

    offset : int, optional
        Starting offset in time (in case of training restart)

    Yields
    ------
    float
        Next noise amount
    """
    offset -= 1
    while True:
        offset += 1
        yield max(m * offset + b, 0)


def _format_helper(x):
    """Format input for logging, depending on type"""
    if isinstance(x, int):
        return "%d" % x
    elif isinstance(x, float):
        return "%.16e" % x
    else:
        return str(x)


def DefaultLoggerFactory(fh, model, maxcols=None, printer=None):
    """Factory function to record likelihood and parameter changes during training

    Parameters
    ----------
    fh : file-like
        Something implementing a ``write()`` method

    model : :class:`~minihmm.hmm.FirstOrderHMM`
        HMM that will be trained

    maxcols : int or None, optional
        If not `None`, only output the first `maxcols` columns of output to `fh`

    printer : file-like or None, optional
        If not `None`, a file-like object to which the first five coluimns of
        output will be dumped (e.g. :obj:`sys.stdout` )


    Returns
    -------
    function
        Logging function for use with :func:`train_baum_welch`
    """
    header = [
        "time",
        "iteration",
        "counted",
        "delta",
        "weighted_logprob",
        "unweighted_logprob",
        "weighted_logprob_per_obs",
        "unweighted_logprob_per_obs",
        "weighted_length",
        "unweighted_length",
    ]
    header += model.get_header()
    if printer is not None:
        printer.write("\t".join(header[:5]))

    header = header[:maxcols]
    fh.write("\t".join(header) + "\n")

    # yapf: disable
    def logfunc(
            model,
            iteration=None,
            counted=None,
            delta=None,
            weighted_logprob=None,
            unweighted_logprob=None,
            weighted_logprob_per_obs=None,
            unweighted_logprob_per_obs=None,
            weighted_length=None,
            unweighted_length=None,
    ):
        ltmp = [
            datetime.datetime.now(),
            iteration,
            counted,
            delta,
            weighted_logprob,
            unweighted_logprob,
            weighted_logprob_per_obs,
            unweighted_logprob_per_obs,
            weighted_length,
            unweighted_length,
        ] + model.get_row()
        ltmp = [_format_helper(X) for X in ltmp]

        # yapf: enable

        fh.write("\t".join(ltmp[:maxcols]))

        if printer is not None:
            printer.write("\t".join(ltmp[:5]))

    return logfunc


#===============================================================================
# INDEX: training functions
#===============================================================================


def bw_worker(
        my_model,
        my_tuple,
        state_prior_estimator=DiscreteStatePriorEstimator(),
        emission_estimator=None,
        transition_estimator=DiscreteTransitionEstimator()
):
    """Collect summary statistics from an observation sequence for Baum-Welch
    training.  In an expectation-maximization context, :py:func:`bw_worker` is
    used in evaluating the Q function in the E step.


    Parameters
    ----------
    my_model : :class:`~minihmm.hmm.FirstOrderHMM`
        Model under which observations are evaluated

    my_tuple : tuple
        Tuple of (:class:`numpy.ndarray`, :class:`float`) corresponding to an
        observation sequence and its weight

    state_prior_estimator : instance of subclass of :class:`~minihmm.estimators.AbstractProbabilityEstimator`, optional
        Estimator that extracts summary statistics regarding probabilities of
        starting in each state. Typically a :class:`~minihmm.estimators.DiscreteStatePriorEstimator`

    transition_estimator : instance of subclass of :class:`~minihmm.estimators.AbstractProbabilityEstimator`, optional
        Estimator that extracts summary statistics regarding transition
        probabilities.  Typically a
        :class:`~minihmm.estimators.DiscreteTransitionEstimator`

    emission_estimator : instance of subclass of :class:`~minihmm.estimators.AbstractProbabilityEstimator`, optional
        Estimator that extracts summary statistics regarding observations
        for each state


    Returns
    -------
    float
        log probability for observation sequence under ``my_model``

    int
        Length of observation sequence

    float
        Weight of observation sequence

    tuple
        numpy.ndarray
            contribution of ``my_obs`` to transition summary statistics

        numpy.array
            contribution of ``my_obs`` to emission summary statistics

        numpy.array
            contribution of ``my_obs`` to state prior summary statistics
    """
    my_obs, my_weight = my_tuple
    obs_logprob, forward, backward, scale_factors, ksi = my_model.forward_backward(my_obs)
    my_A = transition_estimator.reduce_data(
        my_obs, obs_logprob, forward, backward, scale_factors, ksi
    )
    my_E = emission_estimator.reduce_data(
        my_obs, obs_logprob, forward, backward, scale_factors, ksi
    )
    my_pi = state_prior_estimator.reduce_data(
        my_obs, obs_logprob, forward, backward, scale_factors, ksi
    )

    return obs_logprob, len(my_obs), my_weight, (my_A, my_E, my_pi)


#def train_baum_welch2(model,
#                      obs,
#                      state_prior_estimator   = DiscreteStatePriorEstimator(),
#                      transition_estimator    = DiscreteTransitionEstimator(),
#                      emission_estimator      = None,
#                      weights                 = None,
#                      pseudocount_weights     = 0,
#                      observation_weights     = None,
#                      noise_weights           = 0,
#                      log_func                = None,
#                      learning_threshold      = 1e-5,
#                      miniter                 = 10,
#                      maxiter                 = 1000,
#                      start_iteration         = 0,
#                      processes               = 4,
#                      chunksize               = None,
#                      logfile                 = NullWriter(),
#                      printer                 = NullWriter(),
#                      freeze                  = None,
#                     ):
#
#    model_type = type(model)
#    expectation_func = functools.partial(bw_worker,
#                                         model,
#                                         state_prior_estimator = state_prior_estimator,
#                                         emission_estimator    = emission_estimator,
#                                         transition_estimator  = transition_estimator
#                                        )
#
#    def validation_func(obs_stats):
#        """
#        Must return bool
#        """
#        return not any([transition_estimator.is_invalid(obs_stats[0]),
#                        emission_estimator.is_invalid(obs_stats[1]),
#                        state_prior_estimator.is_invalid(obs_stats[2]),
#                       ])
#
#    def maximization_func(obs_stats,
#                          obs_weights,
#                          last_model,
#                          noise_weight        = 0,
#                          pseudocount_weights = None,
#                          pseudocount_model   = None,
#                          freeze              = None):
#        """
#        Must return new model
#        """
#        if freeze is None:
#            freeze = set()
#
#        A_parts, E_Parts, pi_parts = zip(*obs_stats)
#
#        if "state_priors" in freeze:
#            new_state_priors = last_model.state_priors
#        else:
#            new_state_priors = state_prior_estimator.construct_factors(last_model,
#                                                                       pi_parts,
#                                                                       noise_weight,
#                                                                       pseudocounts)
#
#        if "trans_probs" in freeze:
#            new_trans_probs = last_model.trans_probs
#        else:
#            new_trans_probs = transition_estimator.construct_factors(last_model,
#                                                                     A_parts,
#                                                                     noise_weight,
#                                                                     pseudocounts)
#
#        if "emission_probs" in freeze:
#            new_emission_probs = last_model.emission_probs
#        else:
#            new_emission_probs = emission_estimator.construct_factors(model,
#                                                                      E_parts,
#                                                                      noise_weight,
#                                                                      pseudocounts)
#
#        return model_type(state_priors   = new_state_priors,
#                          trans_probs    = new_trans_probs,
#                          emission_probs = new_emission_probs)
#
#    return train_em(model,
#                    obs,
#                    expectation_func  = expectation_func,
#                    validation_func   = validation_func,
#                    maximization_func = maximization_func,
#                    log_func          = log_func,
#
#                    pseudocount_model   = None,
#
#                    pseudocount_weights = pseudocount_weights,
#                    noise_weights       = noise_weights,
#                    observation_weights = observtaion_weights,
#
#                    freeze              = freeze,
#
#                    learning_threshold  = learning_threshold,
#                    miniter             = miniter,
#                    maxiter             = maxiter,
#                    start_iteration     = start_iteration,
#
#                    processes           = processes,
#                    chunksize           = chunksize)
#


# yapf: disable
def train_baum_welch(model,
                     obs,
                     state_prior_estimator   = DiscreteStatePriorEstimator(),
                     transition_estimator    = DiscreteTransitionEstimator(),
                     emission_estimator      = None,
                     pseudocount_weights     = 0.0,
                     noise_weights           = 0.0,
                     observation_weights     = None,
                     learning_threshold      = 1e-5,
                     miniter                 = 10,
                     maxiter                 = 1000,
                     start_iteration         = 0,
                     processes               = 4,
                     chunksize               = None,
                     logfunc                 = None,
                    ):
    """Train an HMM using the Baum-Welch algorithm on one or more unlabeled
    observation sequences.

    After Durbin et al., incorporating equations from solution guide, which
    differ from those in Rabiner

    Parameters
    ----------
    model : :class:`~minihmm.hmm.FirstOrderHMM` or subclass thereof
        Starting HMM, initialized with some set of initial parameters

    obs : list of numpy.ndarray
        One or more observation sequences which will be used for training.

    state_prior_estimator : instance of any subclass of :class:`~minihmm.estimators.AbstractProbabilityEstimator`, optional
        (Default: instance of :class:`~minihmm.estimators.DiscreteStatePriorEstimator`)

    transition_estimator : instance of any subclass of :class:`~minhimm.estimators.AbstractProbabilityEstimator`, optional
        (Default: instance of :class:`~minihmm.estimators.DiscreteTransitionEstimator`)

    emission_estimator : instance of any subclass of AbstractEmissionEstimator

    pseudocount_weights : iterable or number, optional
        Iterator/generator indicating the total weight of pseudocounts that
        should be applied relative to total number of observations in data in
        each cycle of training. How the pseudocounts are distributed is up to
        the Estimators supplied above.

        If an iterator/generator and not infinite, `1e-12` pseudocounts will be
        added after the iterator/generator raises StopIteration. (Default: 0)

    noise_weights : iterable or number, optional
        Iterator/generator specifying the weight of noise relative to total
        number of observations in the internal observation matrices to add to
        each cycle.  How the noise is distributed is up to the Estimators
        supplied above.

        If an iterator/generator and not infinite, no noise will be added after
        the iterator/generator raises  StopIteration.  (Default: 0 (no noise))

    miniter : int or 0, optional
        Minimum number of training cycles (e.g. during which end_delta is
        ignored).  This is useful for allowing added noise to choose suboptimal
        parameters int he hopes of better exploring the parameter space.
        (Default: 100)

    maxiter : int or numpy.inf, optional
        Maximum number of training cycles. Specify numpy.inf to disable this
        termination criterion. (Default: 1000)

    learning_threshold : float, optional
        Minimum change in total log likelihood required to continue training.
        Specify ``-numpy.inf`` to disable this termination criterion (Default:
        1e-5)

    start_iteration : int, optional
        Label for starting iteration (in case training was halted and resumed)

    processes : int, optional
        Number of processes to use during training (default: 4)

    chunksize : int, optional
        Number of observation sequences to send to each process at a time
        (Default: calculated from `len(obs)` and `processes`)

    logfunc : callable, optional
        Logging function. Should accept a model as a first argument, and allow
        arbitrary keywords to pass through.  At present, the following keywords
        are passed: `iteration`, `counted`, `delta`, `weighted_logprob`,
        `unweighted_logprob`, `weighted_logprob_per_obs`,
        `unweighted_logprob_per_obs`, `weighted_length`, `unweighted_length`.


    Returns
    -------
    :class:`~minihmm.hmm.FirstOrderHMM`
        Best (highest-likelihood) model from training trajectory

    :class:`~minihmm.hmm.FirstOrderHMM`
        Final model from training trajectory

    str
        Reason explaining why training ceased. ``MAXITER`` if maximum iterations
        reached. ``CONVERGENCE`` if model converged to `learning_threshold`.
        ``REGRESSION`` if new parameters are worse than old ones, perhaps
        due to noise, but the training iteration is past `miniter`
    """
    # yapf: enable
    if emission_estimator is None:
        raise AssertionError(
            "Emission estimator required by train_baum_welch(), but not supplied"
        )

    oopsie = False

    state_priors = model.state_priors
    emission_factors = model.emission_probs
    trans_probs = model.trans_probs
    model_type = type(model)

    models = []
    logprobs = []
    unweight_logprobs = []

    logprobs_per_length = []
    unweight_logprobs_per_length = []

    last_total_logprob = -numpy.inf
    delta = numpy.inf
    c = start_iteration

    if chunksize is None:
        chunksize = len(obs) // (4 * processes)
    if chunksize < 1:
        chunksize = len(obs) // processes
    if chunksize < 1:
        chunksize = 1

    if isinstance(pseudocount_weights, (int, float)):
        pseudocount_weights = itertools.repeat(1.0 * pseudocount_weights)
    if isinstance(noise_weights, (int, float)):
        noise_weights = itertools.repeat(1.0 * noise_weights)

    if observation_weights is None:
        observation_weights = [1.0] * len(obs)

    weighted_obs = list(zip(obs, observation_weights))

    try:
        while c < maxiter and (delta > learning_threshold or c < miniter) and not oopsie:
            model = model_type(state_priors, emission_factors, trans_probs)
            try:
                noise_weight = next(noise_weights)
            except StopIteration:
                noise_weight = 0

            try:
                pseudocounts = next(pseudocount_weights)
            except StopIteration:
                pseudocounts = 1e-12

            # E-step of Expectation-Maximization
            training_func = functools.partial(
                bw_worker,
                model,
                state_prior_estimator=state_prior_estimator,
                emission_estimator=emission_estimator,
                transition_estimator=transition_estimator
            )
            if processes == 1:
                pool_results = (training_func(X) for X in weighted_obs)
            else:
                pool = multiprocessing.Pool(processes=processes)
                pool_results = pool.map(training_func, weighted_obs, chunksize)
                pool.close()
                pool.join()

            results = []
            anti_results = []
            for my_result in pool_results:
                if any([
                        numpy.isnan(my_result[0]),
                        numpy.isinf(my_result[0]),
                        transition_estimator.is_invalid(my_result[-1][0]),
                        emission_estimator.is_invalid(my_result[-1][1]),
                        state_prior_estimator.is_invalid(my_result[-1][2]),
                ]):
                    anti_results.append(my_result)
                else:
                    results.append(my_result)

            counted = len(results)
            if counted == 0:
                oopsie = True
                break

            obs_logprobs, obs_lengths, obs_weights, obs_stats = zip(*results)
            A_parts, E_parts, pi_parts = zip(*obs_stats)

            obs_weights = numpy.array(obs_weights)
            obs_lengths = numpy.array(obs_lengths)
            obs_logprobs = numpy.array(obs_logprobs)

            weight_logprobs = obs_weights * obs_logprobs
            weight_lengths = obs_weights * obs_lengths

            new_total_logprob = weight_logprobs.sum()
            total_obs_length = weight_lengths.sum()
            logprob_per_obs = new_total_logprob / total_obs_length
            logprobs_per_length.append(logprob_per_obs)

            unweight_total_logprob = obs_logprobs.sum()
            unweight_total_length = obs_lengths.sum()
            unweight_logprob_per_obs = unweight_total_logprob / unweight_total_length
            unweight_logprobs_per_length.append(unweight_logprob_per_obs)

            models.append(model)
            logprobs.append(new_total_logprob)
            unweight_logprobs.append(unweight_total_logprob)

            delta = new_total_logprob - last_total_logprob
            last_total_logprob = new_total_logprob

            if logfunc is not None:
                logfunc(model,
                        iteration = c,
                        counted   = counted,
                        delta     = delta,

                        weighted_logprob   = new_total_logprob,
                        unweighted_logprob = unweight_total_logprob,
                        weighted_logprob_per_obs   = logprob_per_obs,
                        unweighted_logprob_per_obs = unweight_logprob_per_obs,

                        weighted_length   = total_obs_length,
                        unweighted_length = unweight_total_length) # yapf: disable


            # M-step of Expectation-Maximization:
            # Update parameters for next model
            state_priors = state_prior_estimator.construct_factors(
                model, pi_parts, noise_weight=noise_weight, pseudocount_weight=pseudocounts
            )
            trans_probs = transition_estimator.construct_factors(
                model, A_parts, noise_weight=noise_weight, pseudocount_weight=pseudocounts
            )
            emission_factors = emission_estimator.construct_factors(
                model, E_parts, noise_weight=noise_weight, pseudocount_weight=pseudocounts
            )

            c += 1

    except KeyboardInterrupt:
        halted = True

    # report why training stopped
    if c == maxiter:
        reason = "MAXITER"
    elif delta <= learning_threshold:
        if delta >= 0:
            reason = "CONVERGENCE"
        else:
            reason = "REGRESSION"
    elif oopsie == True:
        reason = "ALL_INVALID"
    elif halted == True:
        reason = "HALTED"
    else:
        reason = "SOMETHING TERRIBLE: iter %s, delta %s, counted %s" % (c, delta, counted)

    logprobs = numpy.array(logprobs)
    best_model = models[logprobs.argmax()]

    dtmp = {
        "best_model"                   : best_model,
        "final_model"                  : model,
        "weight_logprobs"              : logprobs,
        "unweight_logprobs"            : unweight_logprobs,
        "weight_logprobs_per_length"   : logprobs_per_length,
        "unweight_logprobs_per_length" : unweight_logprobs_per_length,
        "reason"                       : reason,
        "iterations"                   : c - 1,
        "models"                       : models,
    } # yapf: disable

    return dtmp



#!/usr/bin/env python
"""Pythonic representations of Factors (probability distributions, or, more
generally, affinity functions), which may serve as components of probabilistic
models (e.g. transition or observation probabilities for a
:class:`~minihmm.hmm.FirstOrderHMM` or as models themselves. All probabilistic
models in this library descend from :class:`~minihmm.factors.AbstractFactor`

Factors may be:

    - continuous or discrete
    - univariate or multivariate in their emissions

Factors must be able to:

    - calculate a log probability of an observation via a ``logprob()`` method

    - calculate a probability of an obervation via a ``probability()`` method

    - calculate a log probability of a series of observations via
      a ``get_model_log_likelihood()`` method

    - offer serialization support via pickling (we use :mod:`jsonpickle`
      and, where necessary, implement ``__reduce__()`` methods to make sane
      JSON blobs)

In addition, factors may:
    - supply a ``generate(n)`` function, to generate `n` observations
      from their distribution.

"""
import functools
import copy
import multiprocessing
import numpy
import scipy.stats

from abc import abstractmethod

from minihmm.util import matrix_from_dict, matrix_to_dict

#===============================================================================
# Abstract classes
#===============================================================================


class AbstractFactor(object):
    """Abstract class for all probability distributions
    """

    def probability(self, *args, **kwargs):
        """Return the probability of a single observation

        Parameters
        ----------
        args : list
            List of arguments representing observations

        kwargs : dict
            Dict of arguments representing observations

        Returns
        -------
        float
            Probability of observation
        """
        return numpy.exp(self.logprob(*args, **kwargs))

    def logprob(self, *args, **kwargs):
        """Return the log probability of a single observation

        Parameters
        ----------
        args : list
            List of arguments representing observations

        kwargs : dict
            Dict of arguments representing observations

        Returns
        -------
        float
            Log probability of observation
        """
        return numpy.log(self.probability(*args, **kwargs))

    def get_model_log_likelihood(self, observations, processes=4):
        """Return the log likelihood of a sequence of observations

        Parameters
        ----------
        observations : list-like
            Sequence of observation

        Processes : int, optional
            Number of processes to use when computing likelihood
            (Default: 4)

        Returns
        -------
        float
            Log likelihood of all observations
        """
        if processes == 1:
            return sum((self.logprob(X) for X in observations))
        else:
            pool = multiprocessing.Pool(processes=processes)
            pool_results = pool.map(self.logprob, observations)
            pool.close()
            pool.join()
            return sum(pool_results)

    @abstractmethod
    def generate(self, *args, **kwargs):
        """Sample a random value from the distribution"""

    @abstractmethod
    def get_header(self):
        """Return a list of parameter names corresponding to elements returned
        by `self.get_row()`
        """

    @abstractmethod
    def get_row(self):
        """Serialize parameters as a list, to be used e.g. as a row in a
        :class:`pandas.DataFrame`
        """


#===============================================================================
# Helpers for unpickling / reviving from jsonpickle
#===============================================================================


def _get_scipydistfactor_from_dict(dtmp):
    """Revive a :class:`ScipyDistributionFactor` from a dictionary

    Parameters
    ----------
    dtmp : dict
        Dictionary, created by :meth:`ScipyDistributionFactor._to_dict`

    Returns
    -------
    ScipyDistributionFactor
    """
    # code note: this function cannot be a static method of
    # ScipyDistributionFactor because it is called by its __reduce__()
    # function
    dist_class = getattr(scipy.stats.distributions, dtmp["dist_class"])
    return ScipyDistributionFactor(dist_class(), *dtmp["dist_args"], **dtmp["dist_kwargs"])


#===============================================================================
# Implementable classes
#===============================================================================


class ArrayFactor(AbstractFactor):
    """Univariate probability distribution constructed from a 1D list or array

    Attributes
    ----------
    data : numpy.ndarray
        Table of probabilities indexed by position


    See also
    --------
    MatrixFactor
    """

    def __init__(self, data):
        """Create an ArrayFactor

        Parameters
        ----------
        data : list-like
            Array of probabilities indexed by position
        """
        self.data = numpy.array(data)

    def __eq__(self, other):
        return (self.data == other.data).all()

    def __len__(self):
        return len(self.data)

    def get_header(self):
        """Return a list of parameter names corresponding to elements returned
        by :meth:`ArrayFactor.get_row`
        """
        return [str(X) for X in range(len(self.data))]

    def get_row(self):
        """Serialize parameters as a list, to be used e.g. as a row in a
        :class:`pandas.DataFrame`
        """
        return list(self.data)

    def probability(self, i):
        """Return probability in cell `i` of the array

        Parameters
        ----------
        i : int
            Index of requested probability
        """
        return self.data[i]

    def generate(self):
        """Generate a random sample from the distribution

        Returns
        --------
        int
            Sample generated
        """
        return (self.data.cumsum() >= numpy.random.random()).argmax()


class MatrixFactor(AbstractFactor):
    """Bivariate probability distribution constructed from a two-dimensional
    matrix or array. MatrixFactors can represent joint distributions `P(X, Y)`
    as well as conditional distributions `P(Y|X)`, depending upon whether
    `self.row_conditional` is ``True`` or ``False``.

    Attributes
    ----------
    data : numpy.ndarray
        MxN table of probabilities

    row_conditional : bool, optional
        If ``True``, `data` is a conditional probability table,
        with rows specifying the conditional variable `P(column|row)`
        If ``False``, `data` is athejoint distribution of probabilities
        `P(column, row)`. (Default: ``True``)

    """

    def __init__(self, data, row_conditional=True):
        """Create a MatrixFactor

        Parameters
        ----------
        data : numpy.ndarray
            MxN table of probabilities

        row_conditional : bool, optional
            If ``True``, `data` is a conditional probability table, with rows
            specifying the conditional variable `P(column|row)` If ``False``,
            ``data`` is the joint distribution of probabilities
            `P(column,row)`. (Default: ``True``)
        """
        self.row_conditional = row_conditional
        self.data = numpy.array(data)

    def __eq__(self, other):
        return self.row_conditional == other.row_conditional and (self.data == other.data).all()

    def __len__(self):
        if self.row_conditional is True:
            return self.data.shape[0]
        else:
            return self.data.shape[0] * self.data.shape[1]

    def get_header(self):
        """Return a list of parameter names corresponding to elements returned
        by :meth:`MatrixFactor.get_row`
        """
        shape = self.data.shape
        return ["%d,%d" % (X, Y) for X in range(shape[0]) for Y in range(shape[1])]

    def get_row(self):
        """Serialize parameters as a list, to be used e.g. as a row in a
        :class:`pandas.DataFrame`
        """
        return list(self.data.ravel())

    def probability(self, i, j):
        """Return probability value at `(i, j)` in underlying matrix

        Parameters
        ----------
        i : int
            Row index

        j : int
            Column index

        Returns
        -------
        float
            `P(i|j)` if `self.row_conditional` is ``True``, or
            `P(i,j)` if `self.row_conditional` is ``False``.
        """
        return self.data[i, j]

    def generate(self, i=None, size=1):
        """Sample from the :class:`~minihmm.factors.MatrixFactor`

        Parameters
        ----------
        i : int or None
            Row index, required if `self.row_conditional` is ``True``,
            otherwise ignored

        Returns
        -------
        (int,int)
            Sample, as *(row,col)*
        """
        if self.row_conditional is True:
            #return (self.data[i,:].cumsum() >= numpy.random.random()).argmax()
            return (numpy.random.random() < self.data[i, :].cumsum()).argmax()
        else:
            #TODO: implement!
            pass


class FunctionFactor(AbstractFactor):
    """Probability distribution constructed from functions

    Attributes
    ----------
    _func : function
        Function that evaluates probability

    _generator : function
        Function that samples from the probability distribution

    func_args : list
        Zero or more positional arguments passed to `self._func`
        and `self._generator`

    func_kwargs : dict
        Zero or more keyword arguments passed to `self._func`
        and `self._generator`
    """

    def __init__(self, func, generator_func, *func_args, **func_kwargs):
        """Create a FunctionFactor

        Parameters
        ----------
        func : function
            Function that evaluates probability, given `func_args`,
            `func_kwargs`, and an observation

        generator_func : function
            Function that generates samples, given `func_args`
            and `func_kwargs`

        func_args
            Zero or more positional arguments to pass to `func`
            and `generator func`

        func_kwargs
            Zero or more keyword arguments to pass to `func`
            and `geneator func`

        """
        self.probability = functools.partial(func, *func_args, **func_kwargs)
        self._generator = functools.partial(generator_func, *func_args, **func_kwargs)
        self._func = func
        self._funcname = getattr(self._func, "func_name", getattr(self._func, "__name__", "foo"))
        self._generator_func = generator_func
        self.func_args = copy.deepcopy(func_args)
        self.func_kwargs = copy.deepcopy(func_kwargs)

    def __eq__(self, other):
        return self._func == other._func \
            and self.func_args == other.func_args \
            and self.func_kwargs == other.func_kwargs

    def __repr__(self):
        return "<%s func:'%s'>" % (self.__class__.__name__, self._funcname)

    def __str__(self):
        return repr(self)

    def get_header(self):
        """Return a list of parameter names corresponding to elements returned
        by :meth:`FunctionFactor.get_row`
        """
        ltmp = [str(X) for X in range(len(self.func_args))]
        ltmp += sorted(self.func_kwargs.keys())
        return ltmp

    def get_row(self):
        """Serialize parameters as a list, to be used e.g. as a row in a
        :class:`pandas.DataFrame`
        """
        ltmp = list(self.func_args)
        ltmp += [X[1] for X in sorted(self.func_kwargs.items())]
        return ltmp

    def generate(self, *args, **kwargs):
        """Sample a random value from the distribution

        args
            zero or more positional arguments to pass to pdf or pmf
            *in addition* to those in `self.func_args`

        kwargs
            zero or more keyword arguments to pass to pdf or pmf
            *in addition* to those in `self.func_kwargs`

        Returns
        -------
        object, type depending upon function
            A sample from the distribution
        """
        return self._generator(*args, **kwargs)


class LogFunctionFactor(FunctionFactor):
    """Probability distribution constructed from functions"""

    def __init__(self, func, generator_func, *func_args, **func_kwargs):
        """Create a LogFunctionFactor

        Parameters
        ----------
        func : function
            Function that evaluates log probability, given `func_args`,
            `func_kwargs`, and an observation

        generator_func : function
            Function that generates samples, given `func_args`
            and `func_kwargs`

        func_args
            Zero or more positional arguments to pass to `func` and
            `generator func`

        func_kwargs
            Zero or more keyword arguments to pass to `func` and
            `generator func`

        """
        self.func_args = func_args
        self.func_kwargs = func_kwargs
        self._func = func
        self._funcname = getattr(self._func, "func_name", getattr(self._func, "__name__", "foo"))
        self._generator_func = generator_func
        self.logprob = functools.partial(func, *func_args, **func_kwargs)
        self._generator = functools.partial(generator_func, *func_args, **func_kwargs)


class ScipyDistributionFactor(AbstractFactor):
    """Probability distribution using distributions provided by :mod:`scipy.stats`

    Attributes
    ----------
    distribution : class
        A class describing a distribution from :mod:`scipy.stats`

    Notes
    -----
    These cannot be used in parallelized processes because distributions
    in :mod:`scipy.stats` are not picklable!
    """
    from_dict = staticmethod(_get_scipydistfactor_from_dict)

    def __init__(self, dist_class, *dist_args, **dist_kwargs):
        """Create a ScipyDistributionFactor

        Parameters
        ----------
        dist_class : class
            class for a distribution from :mod:`scipy.stats`
            (e.g. :obj:`scipy.stats.poisson`)

        dist_args
            Zero or more positional arguments to pass to `dist_class`

        dist_kwargs
            Zero or mor keyword arguments to pass to `dist_class`
        """
        self.distribution = dist_class(*dist_args, **dist_kwargs)
        self._dist_class = dist_class
        self.dist_args = dist_args
        self.dist_kwargs = dist_kwargs

        self.prob_fn = self.distribution.pdf
        self.log_prob_fn = self.distribution.logpdf

        # determine whether distribution is continuous
        # or discrete
        try:
            self.distribution.pdf(3)
        except AttributeError:
            self.prob_fn = self.distribution.pmf
            self.log_prob_fn = self.distribution.logpmf

    def __eq__(self, other):
        return self._to_dict() == other._to_dict()

    def get_header(self):
        """Return a list of parameter names corresponding to elements returned
        by :meth:`ScipyDistributionFactor.get_row`

        Returns
        -------
        list
            Names of parameters
        """
        ltmp = [str(X) for X in range(len(self.dist_args))]
        ltmp += sorted(self.dist_kwargs.keys())
        return ltmp

    def get_row(self):
        """Serialize parameters as a list, to be used e.g. as a row in a
        :class:`pandas.DataFrame`

        Returns
        -------
        list
            Values of parameters
        """
        ltmp = list(self.dist_args)
        ltmp += [X[1] for X in sorted(self.dist_kwargs.items())]
        return ltmp

    def _to_dict(self):
        """Convenience function to define pickling/unpickling protocol
        
        Returns
        -------
        dict
            Dictionary of essential information to pickle
        """
        return {
            "model_class": "minihmm.factors.ScipyDistributionFactor",
            "dist_class": self._dist_class.__class__.__name__,
            "dist_args": list(self.dist_args),
            "dist_kwargs": dict(self.dist_kwargs),
        }

    def __reduce__(self):
        return _get_scipydistfactor_from_dict, (self._to_dict(), )

    def __repr__(self):
        return "<%s model:%s parameters:%s>" % (self.__class__.__name__,
                                                 self._dist_class,
                                                 ",".join([str(X) for X in self.dist_args])+\
                                                 ",".join(["%s:%s" % (K, V) for K, V in self.dist_kwargs.items()]) )

    def logprob(self, *args, **kwargs):
        """Return the log probability of observing an observation

        Parameters
        ----------
        args
            zero or more positional arguments to pass to logpdf or logpmf
            *in addition* to those in `self.dist_args`

        kwargs
            zero or more keyword arguments to pass to logpdf or logpmf
            *in addition* to those in `self.dist_kwargs`

        Returns
        -------
        float
            Log probability of observation
        """
        return self.log_prob_fn(*args, **kwargs)

    def probability(self, *args, **kwargs):
        """Return the probability of observing an observation

        args
            zero or more positional arguments to pass to pdf or pmf
            *in addition* to those in `self.dist_args`

        kwargs
            zero or more keyword arguments to pass to pdf or pmf
            *in addition* to those in `self.dist_kwargs`

        Returns
        -------
        float
            Probability of observation
        """
        return self.prob_fn(*args)

    def generate(self, *args, **kwargs):
        """Sample a random value from the distribution"""
        return self.distribution.rvs(*args, **kwargs)



#!/usr/bin/env python
"""Testing functions for FirstOrderHMMs. Viterbi decoding, sequence generation,
the forward-backward algorithm, posterior decoding, and re-training are all tested. 


"""
import unittest
import itertools
import warnings
import pickle

import numpy
import scipy.stats

import jsonpickle
import jsonpickle.ext.numpy
jsonpickle.ext.numpy.register_handlers()

from collections import Counter

from nose.plugins.attrib import attr
from nose.tools import (
    assert_greater_equal,
    assert_less_equal,
    assert_dict_equal,
    assert_list_equal,
    assert_almost_equal,
)
from numpy.testing import (assert_array_equal, assert_array_almost_equal, assert_almost_equal)

from minihmm.hmm import FirstOrderHMM
from minihmm.estimators import (
    UnivariateGaussianEmissionEstimator,
    DiscreteStatePriorEstimator,
    DiscreteEmissionEstimator,
    DiscreteTransitionEstimator,
    PseudocountTransitionEstimator,
)
from minihmm.training import train_baum_welch, neg_exp_noise_gen
from minihmm.factors import (ArrayFactor, MatrixFactor, LogFunctionFactor, ScipyDistributionFactor)

from minihmm.test.common import (
    check_array_equal,
    check_equal,
    get_fourstate_poisson,
)

_FORWARD_SEED = 5139284
_TRAINING_SEED = 134067


class _BaseExample():

    # define these variables in this method in subclasses
    @classmethod
    def do_subclass_setup(cls):

        # name of test suite
        cls.name = ""

        # minimum fraction expected to be equal in viterbi decoding test
        cls.min_frac_equal = 0.8

        # dict representation of hmm
        cls.hmm_dict = None

        # HMM instance used for testing
        cls.generating_hmm = None

    @classmethod
    def setUpClass(cls):
        # Note- 10 is quite slow for multiple tests below, given that the
        # checks use brute-force implementations to verify the dynamic
        # programming implementation used in miniHMM.
        cls.seq_lengths = [5, 10]  #,20]

        cls.test_obs_seqs = []  # test observation sequences
        cls.test_state_seqs = []  # state paths through test observation seqs. not currently used
        cls.decode_tests = []  # tests for Viterbi and posterior decoding

        cls.expected_forward_logprobs = []  # log probabilities for cls.test_obs_seqs
        cls.found_forward_logprobs = []
        cls.found_forward_scalefactors = []
        cls.found_forward_scaled_forward_matrices = []

        cls.expected_joint_logprobs = []  # expected joint probabilities for states and observations

        print("Setting up class %s" % cls.__name__)
        cls.do_subclass_setup()
        hmm = cls.generating_hmm
        numpy.random.seed(_TRAINING_SEED)

        for x in cls.seq_lengths:
            states, obs, _ = hmm.generate(x)
            cls.test_state_seqs.append(states)
            cls.test_obs_seqs.append(obs)

        # precalculate useful quantities on observation sequences
        for obs in cls.test_obs_seqs:
            # run forward algorithm; we'll use their results for tests below
            logprob, scaled_forward, scaled_backward, scale_factors, ksi = hmm.forward_backward(obs)
            cls.found_forward_logprobs.append(logprob)
            cls.found_forward_scalefactors.append(scale_factors)
            cls.found_forward_scaled_forward_matrices.append(scaled_forward)

            # manually calculate probabilities of generated observations via dynamic programming
            # the forward probability of an observation sequence is equal to the sum of its joint
            # probabilities with all possible state paths. We calculate each joint probability here,
            # then sum these to find the total expected probability
            my_states = list(range(hmm.num_states))
            paths = {
                (X, ): hmm.state_priors.logprob(X) + hmm.emission_probs[X].logprob(obs[0])
                for X in my_states
            }

            for my_obs in obs[1:]:
                new_paths = {}
                for partial, prob in paths.items():
                    for new_state in my_states:
                        new_paths[tuple(list(partial) + [new_state])
                                  ] = prob + hmm.trans_probs.logprob(
                                      partial[-1], new_state
                                  ) + hmm.emission_probs[new_state].logprob(my_obs)

                paths = new_paths

            dp_prob = numpy.log(numpy.exp(list(paths.values())).sum())
            cls.expected_forward_logprobs.append(dp_prob)
            cls.expected_joint_logprobs.append(paths)

        # Generate test cases for Viterbi and posterior decoding
        cls.decode_tests = [hmm.generate(1000) for _ in range(10)]

        cls.from_pickle = pickle.loads(pickle.dumps(cls.generating_hmm))
        cls.from_json = jsonpickle.decode(jsonpickle.encode(cls.generating_hmm))

        print("Set up class %s" % cls.__name__)

    # override in subclass
    @unittest.skip
    def test_to_json(self):
        assert False

    def test_from_json(self):
        gen = self.generating_hmm
        rev = self.from_json
        yield check_array_equal, gen.state_priors.data, rev.state_priors.data
        yield check_array_equal, gen.trans_probs.data, rev.trans_probs.data
        for i in range(gen.num_states):
            yield check_equal, gen.emission_probs[i], rev.emission_probs[i]

    @attr("slow")
    def test_logprob(self):
        # make sure fast forward probability calculations match those calced by brute force
        numpy.random.seed(_FORWARD_SEED)
        for n, (obs, expected) in enumerate(zip(self.test_obs_seqs,
                                                self.expected_forward_logprobs)):
            found = self.generating_hmm.fast_forward(obs)
            msg = "Failed fast_forward test case '%s' on HMM '%s'. Expected: '%s'. Found '%s'. Diff: '%s'." % (
                n, self.name, expected, found, abs(expected - found)
            )
            yield assert_almost_equal, expected, found, 7, msg

    def test_fast_forward(self):
        # make sure fast forward probability calculations match those calced by brute force
        numpy.random.seed(_FORWARD_SEED)
        for n, (obs, expected) in enumerate(
            zip(self.test_obs_seqs, self.expected_forward_logprobs)
        ):
            found = self.generating_hmm.fast_forward(obs)
            msg = "Failed fast_forward test case '%s' on HMM '%s'. Expected: '%s'. Found '%s'. Diff: '%s'." % (
                n, self.name, expected, found, abs(expected - found)
            )
            yield assert_almost_equal, expected, found, 7, msg

    def test_forward_logprob(self):
        # make sure vectorized forward probability calculations match those calced by brute force
        numpy.random.seed(_FORWARD_SEED)
        for n, (expected, found) in enumerate(
            zip(self.expected_forward_logprobs, self.found_forward_logprobs)
        ):
            msg = "Failed forward test case '%s' on HMM '%s'. Expected: '%s'. Found '%s'. Diff: '%s'." % (
                n, self.name, expected, found, abs(expected - found)
            )
            yield assert_almost_equal, expected, found, 7, msg

    def test_forward_backward_scaled_forward_set_to_one(self):
        # n.b. at present we're only scaling from timestep 1 onwards
        for n, scaled_forward in enumerate(self.found_forward_scaled_forward_matrices):
            yield assert_almost_equal, scaled_forward[1:].sum(1), 1

    def test_forward_scale_factors_product_sum_is_consistent(self):
        # product of scale factors and forward algorithm at each timestep
        # should equal probability of that sequence up to that point
        #
        # this test is misleading if fast_forward() is wrong

        # n.b. at present we're only scaling from timestep 1 onwards
        for n, obs in enumerate(self.test_obs_seqs):
            _, scaled_forward, scale_factors = self.generating_hmm.forward(obs)
            expected = [
                numpy.exp(self.generating_hmm.fast_forward(obs[:X + 1])) for X in range(len(obs))
            ]
            found = scaled_forward.sum(1) * scale_factors.cumprod()
            yield assert_almost_equal, expected[1:], found[1:]

    def test_forward_backward_logprob(self):
        # test forward algorithm portion of forward_backward
        # make sure forward coefficients match those calculated by brute force
        # at each timestep

        numpy.random.seed(_FORWARD_SEED)
        for n, (obs, expected_logprob) in enumerate(zip(self.test_obs_seqs,
                                                        self.expected_forward_logprobs)):
            (found_logprob, scaled_forward, scaled_backward, scale_factors,
             ksi) = self.generating_hmm.forward_backward(obs)
            msg = "Failed fast_forward test case '%s' on HMM '%s'. Expected: '%s'. Found '%s'. Diff: '%s'." % (
                n, self.name, expected_logprob, found_logprob,
                abs(expected_logprob - found_logprob)
            )
            yield assert_almost_equal, expected_logprob, found_logprob, 7, msg

    def test_forward_backward_scalefactors_product_sum_is_consistent(self):
        # product of scale factors and forward algorithm at each timestep
        # should equal probability of that sequence up to that point
        #
        # this test is misleading if fast_forward() is wrong

        # n.b. at present we're only scaling from timestep 1 onwards
        for n, (obs, scale_factors, scaled_forward) in enumerate(
                zip(self.test_obs_seqs, self.found_forward_scalefactors,
                    self.found_forward_scaled_forward_matrices)):

            expected = [
                numpy.exp(self.generating_hmm.fast_forward(obs[:X + 1])) for X in range(len(obs))
            ]
            found = scaled_forward.sum(1) * scale_factors.cumprod()
            yield assert_almost_equal, expected[1:], found[1:]

    # override in subclass
    @unittest.skip
    def test_forward_backward_backward(self):
        # TODO: test backward component of forward-backward algorithm
        # this is currently tested implicitly, as the tests in test_train.py
        # pass
        assert False

    @attr("slow")
    def test_posterior_decode_path_accuracy(self):
        # make sure posterior decode calls are above accuracy threshold listed above
        for expected_states, obs, _ in self.decode_tests:
            found_states = self.generating_hmm.posterior_decode(obs)[0]
            frac_equal = 1.0 * (expected_states == found_states).sum() / len(expected_states)
            msg = "Failed posterior decode test for test case '%s'. Expected at least %s%% accuracy. Got %s%%." % (
                self.name, self.min_frac_equal, frac_equal
            )
            assert_greater_equal(frac_equal, self.min_frac_equal, msg)

    # override in subclass
    @unittest.skip
    def test_generate(self):
        # FIXME: implement this isolated test.
        #
        #
        # generate() is currently tested implicitly, as retraining on generated
        # sequences in test_train.py actually converges.
        #
        #
        # 1. test length of state and observation sequences is correct
        # 2. test sampling is according to joint distribution
        assert False

    @attr("slow")
    def test_joint_path_logprob(self):
        for n, (obs, expected_joint_probs) in enumerate(zip(self.test_obs_seqs,
                                                            self.expected_joint_logprobs)):
            for path, path_prob in expected_joint_probs.items():
                found_joint_prob = self.generating_hmm.joint_path_logprob(path, obs)
                assert_almost_equal(found_joint_prob, path_prob)

    def test_conditional_path_logprob(self):
        for n, (obs, expected_joint_probs) in enumerate(zip(self.test_obs_seqs,
                                                            self.expected_joint_logprobs)):
            total_logprob = self.generating_hmm.fast_forward(obs)
            for path, path_prob in expected_joint_probs.items():
                found_cond_prob = self.generating_hmm.conditional_path_logprob(path, obs)
                assert_almost_equal(found_cond_prob, path_prob - total_logprob)

    @attr("slow")
    def test_sample(self):
        # Test sampling algorithm by checking the slope and intercept of the regression line
        # between expected and observed numbers of observations for each state path
        num_samples = 10000

        for n, (obs, logprob, joint_probs) in enumerate(zip(
                self.test_obs_seqs, self.expected_forward_logprobs, self.expected_joint_logprobs)):
            cond_probs = {K: V - logprob for K, V in joint_probs.items()}
            paths = self.generating_hmm.sample(obs, num_samples=num_samples)
            samples = Counter([tuple(X.astype(int)) for X in paths])
            expected = numpy.zeros(len(cond_probs))
            found = numpy.zeros(len(cond_probs))

            for n, (k, v) in enumerate(sorted(list(cond_probs.items()))):
                expected[n] = num_samples * numpy.exp(v)
                found[n] = samples.get(k, 0)

            m, b, r, p, std = scipy.stats.linregress(expected, found)
            assert_less_equal(abs(m - 1), 0.02, "Slope '%s' further from 1.0 than expected." % (m))
            assert_less_equal(abs(b), 1.0, "Intercept '%s' further from 0.0 than expected." % (b))
            assert_greater_equal(r, 0.95, "r '%s' less than 0.95 than expected." % r)

    @attr("slow")
    def test_viterbi_path_accuracy(self):
        # make sure viterbi calls are above accuracy threshold listed above
        # with ground-truth states
        for expected_states, obs, _ in self.decode_tests:
            found_states = self.generating_hmm.viterbi(obs)["viterbi_states"]
            frac_equal = 1.0 * (expected_states == found_states).sum() / len(expected_states)
            msg = "Failed viterbi test for test case '%s'. Expected at least %s%% accuracy. Got %s%%." % (
                self.name, self.min_frac_equal, frac_equal
            )
            assert_greater_equal(frac_equal, self.min_frac_equal, msg)


class TestACoin(_BaseExample):
    @classmethod
    def do_subclass_setup(cls):
        cls.name = "Coin example"
        cls.min_frac_equal = 0.69

        cls.generating_hmm = FirstOrderHMM(
            **{
                "state_priors": ArrayFactor([0.005, 0.995]),
                "trans_probs": MatrixFactor(numpy.array([[0.8, 0.2], [0.3, 0.7]])),
                "emission_probs": [ArrayFactor([0.6, 0.4]),
                                   ArrayFactor([0.15, 0.85])],
            }
        )

        cls.hmm_dict = {
            "emission_probs": [],
            "state_priors": {
                "shape": (1, 2),
                "row": [0, 0],
                "col": [0, 1],
                "data": [0.005, 0.995],
            },
            "trans_probs": {
                "shape": (2, 2),
                "row": [0, 0, 1, 1],
                "col": [0, 1, 0, 1],
                "data": [0.8, 0.2, 0.3, 0.7],
            }
        }


class TestBTwoGaussian(_BaseExample):
    @classmethod
    def do_subclass_setup(cls):
        cls.name = "Two gaussian example"
        cls.min_frac_equal = 0.7
        transitions = numpy.matrix([[0.9, 0.1], [0.25, 0.75]])

        cls.generating_hmm = FirstOrderHMM(
            **{
                "trans_probs":
                MatrixFactor(transitions),
                "state_priors":
                ArrayFactor([0.8, 0.2]),
                "emission_probs": [
                    ScipyDistributionFactor(scipy.stats.norm, loc=0, scale=0.5),
                    ScipyDistributionFactor(scipy.stats.norm, loc=5, scale=10)
                ],
            }
        )

        cls.hmm_dict = {
            "emission_probs": [],
            "state_priors": {
                "shape": (1, 2),
                "row": [0, 0],
                "col": [0, 1],
                "data": [0.8, 0.2],
            },
            "trans_probs": {
                "shape": (2, 2),
                "row": [0, 0, 1, 1],
                "col": [0, 1, 0, 1],
                "data": [0.9, 0.1, 0.25, 0.75],
            }
        }


@attr("slow")
class TestCFourPoisson(_BaseExample):

    @classmethod
    def do_subclass_setup(cls):
        cls.name = "Four-state Poisson example"
        cls.min_frac_equal = 0.7
        transitions = numpy.matrix(
            [
                [0.8, 0.05, 0.05, 0.1],
                [0.2, 0.6, 0.1, 0.1],
                [0.01, 0.97, 0.01, 0.01],
                [0.45, 0.01, 0.04, 0.5],
            ]
        )

        cls.generating_hmm = get_fourstate_poisson()
        cls.hmm_dict = {
            "emission_probs": [],
            "state_priors": {
                "shape": (1, 4),
                "row": [0, 0, 0, 0],
                "col": [0, 1, 2, 3],
                "data": [0.7, 0.05, 0.15, 0.10],
            },
            "trans_probs": {
                "shape": (4, 4),
                "row": [0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3],
                "col": [0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3],
                "data": [
                    0.8, 0.05, 0.05, 0.1, 0.2, 0.6, 0.1, 0.1, 0.01, 0.97, 0.01, 0.01, 0.45, 0.01,
                    0.04, 0.5
                ],
            }
        }



#!/usr/bin/env python
"""
"""
import pickle
import unittest


import numpy
import jsonpickle
import jsonpickle.ext.numpy
jsonpickle.ext.numpy.register_handlers()

from nose.tools import assert_equal
from minihmm.test.common import (check_equal, check_array_equal, check_tuple_equal)

from minihmm.factors import (
    ArrayFactor,
    MatrixFactor,
    FunctionFactor,
    LogFunctionFactor,
    ScipyDistributionFactor,
)


class AbstractFactor():

    # list of factors
    factors = []

    # list-of-lists of tuples in each factor's domain
    examples = []

    random_seed = 32853294
    to_generate = 50

    @classmethod
    def setUpClass(cls):
        if hasattr(cls, "do_subclass_setup"):
            cls.do_subclass_setup()

        cls.from_json = []
        cls.from_pickle = []
        cls.generated = []

        for factor in cls.factors:
            cls.from_json.append(jsonpickle.decode(jsonpickle.encode(factor)))
            cls.from_pickle.append(pickle.loads(pickle.dumps(factor)))
            numpy.random.seed(seed=cls.random_seed)
            cls.generated.append([factor.generate() for _ in range(cls.to_generate)])

    # override in subclass
    def test_generate_distribution_is_correct(self):
        assert False

    # override in subclass
    def test_logprob_is_correct(self):
        assert False

    # override in subclass
    def test_probability_is_correct(self):
        assert False

    def test_revive_from_pickle_equal(self):
        # NOTE: only tested if class implements __eq__
        for initial, revived in zip(self.factors, self.from_pickle):
            if hasattr(initial, "__eq__"):
                yield check_equal, initial, revived

    def test_revive_from_json_equal(self):
        # NOTE: only tested if class implements __eq__
        for initial, revived in zip(self.factors, self.from_json):
            if hasattr(initial, "__eq__"):
                yield check_equal, initial, revived

    def test_revive_from_pickle_logprob(self):
        for initial, revived, examples in zip(self.factors, self.from_pickle, self.examples):
            for my_example in examples:
                expected = initial.logprob(*my_example)
                found = revived.logprob(*my_example)
                yield check_equal, expected, found

    def test_revive_from_pickle_prob(self):
        for initial, revived, examples in zip(self.factors, self.from_pickle, self.examples):
            for my_example in examples:
                expected = initial.probability(*my_example)
                found = revived.probability(*my_example)
                yield check_equal, expected, found

    def test_revive_from_json_logprob(self):
        for initial, revived, examples in zip(self.factors, self.from_json, self.examples):
            for my_example in examples:
                expected = initial.logprob(*my_example)
                found = revived.logprob(*my_example)
                yield check_equal, expected, found

    def test_revive_from_json_prob(self):
        for initial, revived, examples in zip(self.factors, self.from_json, self.examples):
            for my_example in examples:
                expected = initial.probability(*my_example)
                found = revived.probability(*my_example)
                yield check_equal, expected, found

    def test_revive_from_pickle_generate(self):
        for initial, revived, generated in zip(self.factors, self.from_pickle, self.generated):
            numpy.random.seed(seed=self.random_seed)
            found = [revived.generate() for _ in range(self.to_generate)]
            yield check_array_equal, generated, found

    def test_revive_from_json_generate(self):
        for initial, revived, generated in zip(self.factors, self.from_json, self.generated):
            numpy.random.seed(seed=self.random_seed)
            found = [revived.generate() for _ in range(self.to_generate)]
            yield check_array_equal, generated, found


#===============================================================================
# Tests for ArrayFactor, MatrixFactor
#===============================================================================


class TestArrayFactor(AbstractFactor):
    @classmethod
    def do_subclass_setup(cls):
        for my_len in range(10, 100, 200):
            ary = numpy.random.random(my_len)
            ary /= ary.sum()
            cls.factors.append(ArrayFactor(ary))
            cls.examples.append([(X, ) for X in numpy.random.randint(0, high=my_len, size=50)])

    def test_revive_from_json_shape(self):
        for initial, revived in zip(self.factors, self.from_json):
            yield check_tuple_equal, initial.data.shape, revived.data.shape

    def test_revive_from_pickle_shape(self):
        for initial, revived in zip(self.factors, self.from_pickle):
            yield check_tuple_equal, initial.data.shape, revived.data.shape

    @unittest.skip
    def test_generate_distribution_is_correct(self):
        assert False

    @unittest.skip
    def test_logprob_is_correct(self):
        assert False

    @unittest.skip
    def test_probability_is_correct(self):
        assert False


class TestMatrixFactor(AbstractFactor):
    @classmethod
    def do_subclass_setup(cls):
        for my_len in range(10, 100, 200):
            ary = numpy.random.random((my_len, my_len))
            ary = (ary.T / ary.sum(1)).T
            cls.factors.append(MatrixFactor(ary, row_conditional=True))
            cls.examples.append(
                [(X, Y) for (X, Y) in numpy.random.randint(0, high=my_len, size=(50, 2))]
            )

            cls.factors.append(MatrixFactor(ary, row_conditional=False))
            cls.examples.append(
                [(X, Y) for (X, Y) in numpy.random.randint(0, high=my_len, size=(50, 2))]
            )

    def test_revive_from_json_shape(self):
        for initial, revived in zip(self.factors, self.from_json):
            yield check_tuple_equal, initial.data.shape, revived.data.shape

    def test_revive_from_pickle_shape(self):
        for initial, revived in zip(self.factors, self.from_pickle):
            yield check_tuple_equal, initial.data.shape, revived.data.shape

    @unittest.skip
    def test_generate_distribution_is_correct(self):
        assert False

    @unittest.skip
    def test_logprob_is_correct(self):
        assert False

    @unittest.skip
    def test_probability_is_correct(self):
        assert False


#===============================================================================
# Tests for ScipyDistributionFactor
#===============================================================================

from scipy.stats.distributions import binom, chi2, norm


class TestScipyDistributionFactor(AbstractFactor):

    factors = [
        ScipyDistributionFactor(binom, n=5, p=0.15),
        ScipyDistributionFactor(chi2, df=3),
        ScipyDistributionFactor(norm, loc=0.5, scale=2),
    ]

    examples = [
        [(X, ) for X in numpy.arange(6)],
        [(X, ) for X in numpy.linspace(0, 20, 100)],
        [(X, ) for X in numpy.linspace(-5, 5, 30)],
    ]

    @unittest.skip
    def test_generate_distribution_is_correct(self):
        assert False

    @unittest.skip
    def test_logprob_is_correct(self):
        assert False

    @unittest.skip
    def test_probability_is_correct(self):
        assert False


#===============================================================================
# Tests for FunctionFactor, LogFunctionFactor
#===============================================================================


def _fact_func1(x):
    return 1.0 / numpy.exp(-x)


def _fact_func2(x, **kwargs):
    return x + sum(kwargs.values())


def _gen_func(*args, **kwargs):
    return 5


class TestFunctionFactor(AbstractFactor):

    factors = [
        FunctionFactor(_fact_func1, _gen_func),
        FunctionFactor(_fact_func2, _gen_func, a=5, b=2)
    ]

    examples = [
        [(X, ) for X in numpy.linspace(1, 100, 10)],
        [(X, ) for X in numpy.linspace(1, 100, 10)],
    ]

    @unittest.skip
    def test_generate_distribution_is_correct(self):
        assert False

    @unittest.skip
    def test_logprob_is_correct(self):
        assert False

    @unittest.skip
    def test_probability_is_correct(self):
        assert False


class TestLogFunctionFactor(AbstractFactor):
    factors = [
        LogFunctionFactor(_fact_func1, _gen_func),
        LogFunctionFactor(_fact_func2, _gen_func, a=5, b=2)
    ]

    examples = [
        [(X, ) for X in numpy.linspace(0, 1, 10)],
        [(X, ) for X in numpy.linspace(0, 1, 10)],
    ]

    @unittest.skip
    def test_generate_distribution_is_correct(self):
        assert False

    @unittest.skip
    def test_logprob_is_correct(self):
        assert False

    @unittest.skip
    def test_probability_is_correct(self):
        assert False





#!/usr/bin/env python
"""
"""
import unittest

# At present, we have no isolated tests for estimators. However, some estimators
# are tested in end-to-end tests in training cycles in test_training.py
@unittest.skip
def test_placeholder():
    assert False



#!/usr/bin/env python
"""Contains classes that represent first-order Hidden Markov Models. These
encapsulate methods for:

    #. Assigning state labels to sequences of observations, decoding via
       the Viterbi algorithm or by posterior decoding.

    #. Computing the probability of observing a sequence of observations via
       the forward algorithm

    #. Generating sequences of observations

All classes here support multivariate and univariate emissions (observation
sequences) that can be continuous or discrete.

Training utilities for estimating model parameters may be found in
:mod:`minihmm.training`. Utilities for manipulating higher-order models may be
found in :mod:`minihmm.represent`



References
----------
[Durbin1998]
    Durbin R et al. (1998). Biological sequence analysis: Probabilistic models
    of proteins and nucleic acids. Cambridge University Press, New York.
    ISBN 978-0-521-62971-3

[Rabiner1989]
    Rabiner, LR (1989). A Tutorial on Hidden Markov Models and Selected
    Applications in Speech Recognition. Proceedings of the IEEE, 77(2), pp
    257-286

[WikipediaForwardBackward]
    http://en.wikipedia.org/wiki/Forward%E2%80%93backward_algorithm

[WikipediaViterbi]
    http://en.wikipedia.org/wiki/Viterbi_algorithm
"""
import warnings
import copy
import numpy
import jsonpickle
import jsonpickle.ext.numpy
jsonpickle.ext.numpy.register_handlers()

from minihmm.factors import (
    AbstractFactor,
    ArrayFactor,
    MatrixFactor,
    FunctionFactor,
    LogFunctionFactor,
)


class FirstOrderHMM(AbstractFactor):
    """First-order homogeneous hidden Markov model.

    Observations/emissions can be multivariate
    """

    def __init__(self, state_priors=None, emission_probs=None, trans_probs=None):
        """Create a First order hidden Markov model

        Parameters
        ----------
        state_priors : :class:`~minihmm.factors.ArrayFactor`
            Probabilities of starting in any state

        emission_probs  : list of Factors
            Probability distributions describing the probabilities of observing
            any emission in each state. If a list, the types of factors need
            not be identical (e.g. some could be Gaussian, others T-distributed,
            et c)

        trans_probs : :class:`~minihmm.factors.MatrixFactor`
            Matrix describing transition probabilities from each state (first
            index) to each other state (second index).
        """
        splen = len(state_priors)
        tlen = len(trans_probs)
        elen = len(emission_probs)
        if splen != tlen or splen != elen:
            raise ValueError(
                "State priors, transition probabilities, and emission factors "
                "must have same length. Found %s, %s, %s instead."
                % (splen, tlen, elen)
            )

        self.num_states = len(state_priors)
        self.state_priors = state_priors
        self.emission_probs = emission_probs
        self.trans_probs = trans_probs
        self.__logt = None

    @property
    def _logt(self):
        """Log of transition probability matrix"""
        # lazily evaluate log of transition probabilities on first use
        if self.__logt is None:
            with warnings.catch_warnings():
                warnings.filterwarnings(
                    "ignore", "divide by zero encountered in log", RuntimeWarning
                )
                self.__logt = numpy.log(self.trans_probs.data)

        return self.__logt

    @property
    def trans_probs(self):
        return self._trans_probs

    @trans_probs.setter
    def trans_probs(self, value):
        self._trans_probs = value
        # maintain synchrony with log form used in many calculations
        self.__logt = None

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return "<%s, %s states>" % (self.__class__.__name__, self.num_states)

    def get_header(self):
        """Return a list of parameter names corresponding to elements returned
        by :meth:`FirstOrderHMM.get_row`

        Returns
        -------
        list
            List of parameter names
        """
        ltmp = ["sp_%s" % X for X in self.state_priors.get_header()]
        ltmp += ["t_%s" % X for X in self.trans_probs.get_header()]
        for n, e_prob in enumerate(self.emission_probs):
            ltmp += ["e%d_%s" % (n, X) for X in e_prob.get_header()]

        return ltmp

    def get_row(self):
        """Serialize parameters as a list, to be used e.g. as a row in a
        :class:`pandas.DataFrame`

        Returns
        -------
        list
            List of parameter values
        """
        ltmp = self.state_priors.get_row()
        ltmp += self.trans_probs.get_row()
        for e_prob in self.emission_probs:
            ltmp += e_prob.get_row()

        return ltmp

    def to_json(self):
        """Return a string JSON blob encoding `self`"""
        using_funcfactor = False
        for factor in self.emission_probs:
            if isinstance(factor, (FunctionFactor, LogFunctionFactor)):
                warnings.warn(
                    "(Log)FudenctionFactors may only be successfully revived "
                    "from JSON if defined in a model, or if revived before "
                    "scope is lost",
                    UserWarning
                )

        return jsonpickle.encode(self)

    @staticmethod
    def from_json(stmp):
        """Revive a model from a JSON blob

        Parameters
        ----------
        stmp : str
            JSON blob encoding an HMM

        Returns
        -------
        :class:`FirstOrderHMM`
        """
        return jsonpickle.decode(stmp)

    def probability(self, emission):
        """Compute the probability of observing a sequence of emissions. This
        number is likely to undeflow for long sequences.

        Parameters
        ----------
        emissions : numpy.ndarray
            Sequence of observations

        Returns
        -------
        float
            Probability of sequence of emissions
        """
        return numpy.exp(self.logprob(emission))

    def logprob(self, emission):
        """Compute the log probability of observing a sequence of emissions.

        Parameters
        ----------
        emissions : numpy.ndarray
            Sequence of observations

        Returns
        -------
        float
            log probability of sequence of emissions
        """
        return self.fast_forward(emission)

    def fast_forward(self, emissions):
        """Compute the log probability of observing a sequence of emissions.

        More memory efficient implementation of the forward algorithm,
        retaining only the probability of the terminal and penultimate states
        at each step.  This implementation is not useful for posterior
        decoding, which requires the probability of all intermediate states.
        For that purpose, an alternate implementation is provided by
        :meth:`FirstOrderHMM.forward`

        Numerical underflows are prevented by scaling probabilities at each
        step, following the procedure given in Rabiner (1989), "A Tutorial on
        Hidden Markov Models and Selected Applications in Speech Recognition"

        Parameters
        ----------
        emissions : numpy.ndarray
            Sequence of observations

        Returns
        -------
        float
            log probability of sequence of emissions
        """
        log_probability = 0
        T = self.trans_probs.data
        O0 = numpy.diag(
            [self.emission_probs[X].probability(emissions[0]) for X in range(self.num_states)]
        )
        prev_states_scaled = self.state_priors.data.dot(O0)

        for t in range(1, len(emissions)):
            Ot = numpy.diag(
                [self.emission_probs[X].probability(emissions[t]) for X in range(self.num_states)]
            )
            f = prev_states_scaled.dot(T.dot(Ot))
            c = f.sum()
            prev_states_scaled = f / c
            log_probability += numpy.log(c)

        return log_probability

    def forward(self, emissions):
        """Calculates the log-probability of observing a sequence of emissions,
        regardless of the state sequence, using the Forward Algorithm. This
        implementation also retains intermediate state information useful in
        posterior decoding or Baum-Welch training.

        Numerical underflows are prevented by scaling probabilities at each step,
        following the procedure given in Rabiner (1989), "A Tutorial on Hidden
        Markov Models and Selected Applications in Speech Recognition"

        Vectorized implementation from Wikipedia (2014-02-20):
        http://en.wikipedia.org/wiki/Forward%E2%80%93backward_algorithm

        Parameters
        ----------
        emissions : numpy.ndarray
            Sequence of observations

        Returns
        -------
        float
            log probability of sequence of emissions

        numpy.ndarray
            `[time x num_states]` Array representing  scaled forward algorithm
            vector, indicating the forward probability of being in each state
            at time `T`, given the model and the observation trajectory from
            `t = 0` to `t = T`

        numpy.ndarray
            `[time x 1]` Array of scaling constants used at each step as
            described in Rabiner 1989.  The sum of the log of these equals the
            log probability of the observation sequence.
        """
        total_logprob, scaled_forward, _, scale_factors, _ = self.forward_backward(emissions, calc_backward=False)
        return total_logprob, scaled_forward, scale_factors

    def forward_backward(self, emissions, calc_backward=True):
        """Calculates the forward algorithm, the backward algorithm, and
        sufficient statistics useful in Baum-Welch calculations, all in
        factored and vectorized forms.

        Numerical underflows are prevented by scaling forward  probabilities at
        each step, following the procedure given in Rabiner (1989), "A Tutorial
        on Hidden Markov Models and Selected Applications in Speech
        Recognition"

        Vectorized implementation adapted from Wikipedia (2014-02-20):
        http://en.wikipedia.org/wiki/Forward%E2%80%93backward_algorithm but
        using scaling as described in Rabiner1989 and Durbin1998, rather than
        the Wikipedia article


        Parameters
        ----------
        emissions : numpy.ndarray
            Observations

        calc_backward : bool, optional
            If True, perform backward pass and calculate Rabiner's ksi table


        Returns
        -------
        float
            log probability of sequence of emissions

        numpy.array
            Scaled forward algorithm vector of dim `[time x num_states]`,
            indicating the forward probability of being in each state at time
            `T`, given the model and the observation trajectory from `t = 0` to
            `t = T`

        numpy.ndarray
            Scaled backward algorithm vector of dim `[time x num_states]`,
            indicating the backward probability of being in each state at time
            `T`, given the model and the observation trajectory from `t = T` to
            `t = end`

        numpy.ndarray
            `[time x 1]` Array of scaling constants used at each step as
            described in Rabiner 1989.  The sum of the log of these equals the
            log probability of the observation sequence.

        numpy.ndarray
            `[time x num_states x num_states]` ksi table, as described in
            Rabiner 1989.  At each time `t`, `ksi[t,i,j]` gives the posterior
            probability of transitioning from state `i` to state `j`. From this
            table it is trivial to derive the expected number of transitions
            from state `i` to state `j`, or the posterior probability of being
            in state `i` or `j` at a given timepoint, by taking the appropriate
            sum.


        Notes
        -----
        This implementation casts everything to ``numpy.float128``. Whether
        this will actually force use of IEEE float128 depends on local C
        library implementations
        """
        # probability sequence indexed by timeslice. columns are end states
        scaled_forward = numpy.tile(numpy.nan, (len(emissions), self.num_states))
        scale_factors = numpy.ones(len(emissions))
        T = self.trans_probs.data
        O = []

        # initialize as prior + likelihood of emissions
        O.append([self.emission_probs[X].probability(emissions[0]) for X in range(self.num_states)])
        scaled_forward[0, :] = self.state_priors.data.dot(numpy.diag(O[0]))

        # can get underflows here from very improbable emissions
        #
        # can get nan for probability if f.sum() is 0, in other words, if a
        # given observation is very improbable for all models and underflows
        # for all models, then c = 0, and f/c = [nan,nan,...,nan]
        #
        # This then forces all future probabilities to be set to nan,
        # which messes up forward and backward calculations.
        # In this case, using HighPrecisionFirstOrderHMM will work,
        # but at a cost for speed
        for t in range(1, len(emissions)):
            O.append(
                [self.emission_probs[X].probability(emissions[t]) for X in range(self.num_states)]
            )
            f = scaled_forward[t - 1, :].dot(T.dot(numpy.diag(O[t])))
            c = f.sum()
            scaled_forward[t, :] = f / c
            scale_factors[t] = c

        if calc_backward is True:
            # backward calc
            scaled_backward = numpy.zeros((len(emissions), self.num_states))
            scaled_backward[-1, :] = 1.0 / scale_factors[
                -1
            ]  # <---- Wikipedia says not to scale final timestep; Rabiner & Durbin say to
            for t in range(len(emissions) - 1)[::-1]:
                scaled_backward[t, :] = T.dot(numpy.diag(O[t + 1]).dot(scaled_backward[t + 1, :])
                                              ) / scale_factors[t]

            # ksi calc
            # NOTE: this is a complete calculation despite the fact that we
            #       are working in a scaled space, because the scale factors
            #       end up equaling 1/P(O|lambda), which means we can compute
            #       in scaled space and get an unscaled result if we do not
            #       divide by P(O|lambda)
            O = numpy.array(O)
            ksi = scaled_forward[:-1, :, None] * scaled_backward[1:, None, :] * T[None, :, :
                                                                                  ] * O[1:, None, :]

        else:
            scaled_backward = None
            ksi = None

        if numpy.isnan(scale_factors).any():
            total_logprob = -numpy.Inf
        else:
            total_logprob = numpy.log(scale_factors).sum()

        return total_logprob, scaled_forward, scaled_backward, scale_factors, ksi

    def posterior_decode(self, emissions):
        """Find the most probable state for each individual state in the
        sequence of emissions, using posterior decoding. Note, this objective
        is distinct from finding the most probable sequence of states for all
        emissions, as is given in Viterbi decoding. This alternative may be
        more appropriate when multiple paths have similar probabilities.

        Parameters
        ----------
        emissions : numpy.ndarray
            Sequence of observations


        Returns
        -------
        numpy.ndarray
            An array of dimension `[t x 1]` of the most likely states at each
            point `t`

        numpy.ndarray
            An array of dimension `[t x k]` of the posterior probability of
            being in state `k` at time `t`
        """
        _, forward, backward, scale_factors, _ = self.forward_backward(emissions)
        posterior_probs = forward * backward * scale_factors[:, None]
        most_likely_states = (posterior_probs).argmax(1)

        return most_likely_states, posterior_probs

    def generate(self, length):
        """Generates a random sequence of states and emissions from the HMM

        Parameters
        ----------
        length : int
            Length of sequence to generate


        Returns
        -------
        numpy.ndarray
            Array of dimension `[t x 1]` indicating the HMM state at each
            timestep

        numpy.ndarray
            Array of dimension `[t x Q]` indicating the observation at each
            timestep.  `Q = 1` for univariate HMMs, or more than 1 if
            observations are multivariate.

        float
            Joint log probability of generated state and observation sequence.
            **Note**: this is different from the log probability of the
            observation sequence alone, which would be the sum of its joint
            probabilities with all possible state sequences.


        Notes
        -----
        The HMM can only generate sequences if all of its EmissionFactors are
        generative. I.e. if using :class:`~minihmm.factors.FunctionFactor` or
        :class:`~minihmm.factors.LogFunctionFactor`, generator functions must
        be specified at their instantiation.
        """
        states = []
        emissions = []

        states.append(self.state_priors.generate())
        emissions.append(self.emission_probs[states[0]].generate())

        logprob = self.state_priors.logprob(states[0])
        logprob += self.emission_probs[states[0]].logprob(emissions[0])

        for i in range(1, length):
            new_state = self.trans_probs.generate(states[i - 1])
            new_obs = self.emission_probs[new_state].generate()

            states.append(new_state)
            emissions.append(new_obs)

            logprob += self.trans_probs.logprob(states[-1], new_state)
            logprob += self.emission_probs[new_state].logprob(new_obs)

        return numpy.array(states).astype(int), numpy.array(emissions), logprob

    def joint_path_logprob(self, path, emissions):
        """Return log P(path, emissions) evaluated under this model

        Parameters
        ----------
        path : list-like
            Sequence of states

        emissions : list-like
            Sequence of observations

        Returns
        -------
        float
            Log probability of P(path, emissions)
        """
        T = self._logt
        joint_logprob = self.state_priors.logprob(path[0])
        joint_logprob += self.emission_probs[path[0]].logprob(emissions[0])

        for i in range(1, len(emissions)):
            joint_logprob += T[path[i - 1], path[i]] + self.emission_probs[path[i]].logprob(
                emissions[i]
            )

        return joint_logprob

    def conditional_path_logprob(self, path, emissions):
        """Return log P(path | emissions) evaluated under this model

        Parameters
        ----------
        path : list-like
            Sequence of states

        emissions : list-like
            Sequence of observations

        Returns
        -------
        float
            Log probability of `P(path | emissions)`
        """
        # calculate conditional pat logprob as P(path, emissions) - P(emissions)
        # P(emissions) from fast_forward
        obs_logprob = self.fast_forward(emissions)
        joint_logprob = self.joint_path_logprob(path, emissions)

        return joint_logprob - obs_logprob

    def sample(self, emissions, num_samples=1):
        """Sample state sequences from the distribution `P(states | emissions)`,
        by tracing backward through the matrix of forward probabilities.
        See Durbin1997 ch 4.3, section "Probabilistic sampling of aligments"

        Parameters
        ----------
        emissions : numpy.ndarray
            Sequence of observations

        num_samples : int, optional
            Number of state paths to generate (Default: 1)


        Returns
        -------
        list
            List of state paths
        """
        L = len(emissions)
        T = self.trans_probs.data

        total_logprob, scaled_forward, _, scale_factors, _ = self.forward_backward(
            emissions, calc_backward=False
        )
        randos = numpy.random.random(size=(num_samples, L))
        final_state_cumsums = scaled_forward[-1, :].cumsum()

        paths = []
        for n in range(num_samples):
            my_path = numpy.full(len(emissions), -1, dtype=int)

            # because probabilty at all steps is scaled to one, we can just
            # examine cumsum of final step to start
            last_state = final_state_cumsums.searchsorted(randos[n, 0], side="right")
            #last_state = (scaled_forward[-1,:].cumsum() >= randos[n,0]).argmax()
            my_path[-1] = last_state

            for i in range(1, L):
                pvec = T[:, last_state] \
                       * self.emission_probs[last_state].probability(emissions[-i]) \
                       * scaled_forward[-i-1, :] \
                       / scaled_forward[-i, last_state] \
                       / scale_factors[-i]

                #last_state = (pvec.cumsum() >= randos[n, i]).argmax()
                last_state = pvec.cumsum().searchsorted(randos[n, i], side="right")
                my_path[-i - 1] = last_state

            paths.append(my_path)

        return paths


#     commented out - this actually samples according to joint distribution P(states,obs), NOT P(states|obs)
#     def sample(self, emissions):
#         """Probabilistically sample state sequences from the joint distribution
#         P(emissions, states), using a modified Viterbi algorithm
#
#         See Durbin1997 ch 4.3
#
#         Parameters
#         ----------
#         emissions : numpy.ndarray
#             Sequence of observations
#
#         Returns
#         -------
#         dict
#             Results of decoding, with the following keys:
#
#             `states`
#                 :class:`numpy.ndarray`. Decoded labels for each position in
#                 emissions[start:end]
#
#             `logprob`
#                 :class:`float`. Joint log probability of sequence of
#                 `emissions[start:end]` and the decoded state sequence
#         """
#         rvals = numpy.random.random(len(emissions))
#         T     = self._logt
#
#         prev_probs = numpy.array([self.state_priors.logprob(X) + self.emission_probs[X].logprob(emissions[0])\
#                                   for X in range(self.num_states)])
#
#         newstate   = (prev_probs.cumsum() >= rvals[0]).argmax()
#         state_path = [newstate]
#         logprob    = prev_probs[newstate]
#
#         for r, x in zip(rvals, emissions)[1:]:
#             emission_logprobs = numpy.array([X.logprob(x) for X in self.emission_probs])
#             current_probs     = T[state_path[-1],:] + emission_logprobs + logprob
#             newstate          = (current_probs.cumsum() >= r).argmax()
#
#             state_path.append(newstate)
#             logprob += current_probs[newstate]
#
#         dtmp = {
#             "sampled_states"  : state_path,
#             "logprob"         : logprob,
#         }
#         return dtmp

    def viterbi(self, emissions, start=0, end=None):
        """Finds the most likely state sequence underlying a set of emissions
        using the Viterbi algorithm. Also returns the natural log probability
        of that state sequence.

        See http://en.wikipedia.org/wiki/Viterbi_algorithm

        Parameters
        ----------
        emissions : numpy.ndarray
            Sequence of observations

        start : int, optional
            Starting point in emissions to consider (Default: 0)

        end : int, optional
            Ending point in emissions to consider (Default: None, end of sequence)


        Returns
        -------
        dict
            Results of decoding, with the following keys:

            `viterbi_states`
                :class:`numpy.ndarray`. Decoded labels for each position in
                emissions[start:end]

            `logprob`
                :class:`float`. Joint log probability of sequence of
                `emissions[start:end]` and the decoded state sequence

            `state_paths`
                dict[state] = log probability of final symbol being in that state
        """
        state_dict = {K: [K] for K in range(self.num_states)}
        prev_probs = numpy.array([self.state_priors.logprob(X) + self.emission_probs[X].logprob(emissions[start]) \
                                  for X in range(self.num_states)])

        T = self._logt
        for x in emissions[start + 1:end]:
            new_state_dict = {}
            emission_logprobs = numpy.array([X.logprob(x) for X in self.emission_probs])
            current_probs = T + emission_logprobs[None, :] + prev_probs[:, None]

            new_probs = current_probs.max(0)
            new_state_dict = { X : state_dict[current_probs[:, X].argmax()] + [X] \
                                   for X in range(self.num_states) }

            prev_probs = new_probs
            state_dict = new_state_dict

        final_label = prev_probs.argmax()
        total_logprob = prev_probs.max()
        states = numpy.array(state_dict[final_label])

        dtmp = {
            "viterbi_states": states,
            "state_paths": state_dict,
            "logprob": total_logprob,
        }
        return dtmp



#!/usr/bin/env python
"""Estimator classes for Baum-Welch training. Estimators determine how
parameters for the HMM are re-estimated from observations during each training
cycle.

Estimators must be able to:

    1. Determine whether a set of observations is valid under the probability
       distribution. This is implemented by the methods ``is_valid()`` and
      ``is_invalid()``

    2. Reduce a set of observations to expected summary statistics (expectation
       step), implemented by the method ``reduce_data()``

    3. Estimate improved parameters for the model from the expected statistics
    (maximization step), and, using these parameters, initialize new Factors.
    These steps are implemented by the method ``construct_factors()``

Here we provide examples of estimators for discrete and continuous
distributions, as well as "frozen" estimators which keep parameter values
constant through Baum-Welch training.
"""

from minihmm.factors import (
    ArrayFactor,
    MatrixFactor,
    FunctionFactor,
    LogFunctionFactor,
    ScipyDistributionFactor,
)

from abc import abstractmethod
import numpy
import scipy.stats
import copy

#===============================================================================
# INDEX: helper functions
#===============================================================================


def get_model_noise(template, weight, assymetric_weights=1):
    """Create an array of noise,  with total number of counts equal to
    `weight*template.sum()` and shape equal to template.shape

    Parameters
    ----------
    template : numpy.ndarray
        Array indicating where to put noise, and in what relative distributions

    weight : float
        Relative weight of noise relative to data

    pseudocount_weights : float or numpy.ndarray
        If given, assymetric weights of pseudocounts to noise (e.g. set cells
        to zero to prevent noise from being added to those regions)

    Returns
    -------
    :class:`numpy.ndarray`
        Numpy array containing noise
    """
    noise = numpy.random.random(template.shape) * assymetric_weights
    noise = noise / noise.sum() * weight * template.sum()
    return noise


#===============================================================================
# INDEX: estimators
#===============================================================================


class AbstractProbabilityEstimator(object):
    """Helper class for reestimation of probabilities in Baum-Welch training.
    Subclasses will be used during training to extract sufficient statistics
    from observation data (via calls to
    :meth:`AbstractProbabilityEstimator.reduce_data`, and to create new factors
    from those statistics (via calls to
    :meth:`AbstractProbabilityEstimator.construct_factors`)
    """

    @abstractmethod
    def is_valid(self, reduced_data):
        """Return true if data is valid and should be included in current round
        of reestimation

        Parameters
        ----------
        reduced_data
            output from :meth:`AbstractProbabilityEstimator.reduce_data`
        """
        pass

    def is_invalid(self, reduced_data):
        """Return true if data is invalid and should be excluded from current
        round of reestimation

        Parameters
        ----------
        reduced_data
            output from :meth:`AbstractProbabilityEstimator.reduce_data``
        """
        return not self.is_valid(reduced_data)

    @abstractmethod
    def reduce_data(self, my_obs, obs_logprob, forward, backward, scale_factors, ksi):
        """Collect data from a single observation sequence and reduce it to a form
        amenable for factor construction by
        :meth:`AbstractProbabilityEstimator.construct_factors`

        Parameters
        ----------
        my_obs : list-like
            Observation sequence

        obs_logprob : float
            Observation logprob, from
            :meth:`~minihmm.hmm.FirstOrderHMM.forward_backward`

        forward : numpy.ndarray
            Scaled forward probability matrix, from
            :meth:`minihmm.hmm.FirstOrderHMM.forward_backward`

        backward : numpy.ndarray
            Scaled backward probability matrix, from
            :meth:`minihmm.hmm.FirstOrderHMM.forward_backward`

        scale_factors : numpy.ndarray
            Scale factors used in scaling `forward` and `backward` to
            floating-point friendly sizes, from
            :meth:`minihmm.hmm.FirstOrderHMM.forward_backward`

        ksi : numpy.ndarray
            `MxNxT` array describing the full probability of being in state `M`
            at time `t` and state `N` at time `t+1`. From
            :meth:`minihmm.hmm.FirstOrderHMM.forward_backward`

        Returns
        -------
        numpy.ndarray
            Array of appropirate shape containing sufficient statistics
        """
        pass

    @abstractmethod
    def construct_factors(self, model, reduced_data, noise_weight=0, pseudocount_weight=1e-10):
        """Construct factor(s) for an HMM using reduced data from observation
        sequences

        Parameters
        ----------
        model : :class:`~minihmm.hmm.FirstOrderHMM` or subclass

        reduced_data : numpy.ndarray
            sufficient statistics for observations, from
            :meth:`AbstractProbabilityEstimator.reduce_data`

        noise_weight : float, optional
            weight of noise to add, relative to number of of observations
            (e.g. transition counts, state prior counts, emission counts, et c)
            in data set. (Default: 0)

        pseudocount_weight : float, optional
            weight of pseudocounts to add, relative to number of of observations
            (transition counts, state prior counts, emission counts, et c)
            in data set (Default: 1e-8)

        Returns
        -------
        some sort of probability factor
        """
        pass


class _FrozenParameterEstimator(AbstractProbabilityEstimator):
    """Keep some parameter constant, completely ignoring observations
    """

    def is_valid(self, reduced_data):
        return True

    def reduce_data(self, my_obs, obs_logprob, forward, backward, scale_factors, ksi):
        """Completely ignore data,  since parameter is frozen"""
        return None


class FrozenStatePriorEstimator(_FrozenParameterEstimator):
    """Keep state priors constant, completely ignoring observation data
    """

    def construct_factors(self, model, reduced_data, noise_weight=0, pseudocount_weight=1e-10):
        """Return model's previous state prior factor"""
        return model.state_priors


class FrozenTransitionEstimator(_FrozenParameterEstimator):
    """Keep transition estimates constant,  completely ignoring observation data
    """

    def construct_factors(self, model, reduced_data, noise_weight=0, pseudocount_weight=1e-10):
        """Return model's previous transition factor"""
        return model.trans_probs


class FrozenEmissionEstimator(_FrozenParameterEstimator):
    """Keep emission estimates constant,  completely ignoring observation data
    """

    def construct_factors(self, model, reduced_data, noise_weight=0, pseudocount_weight=1e-10):
        """Return model's previous emission factors"""
        return model.emission_probs


class _DiscreteParameterEstimator(AbstractProbabilityEstimator):
    """Estimator that models CPDs as tables
    """

    def is_valid(self, reduced_data):
        return not (numpy.isnan(reduced_data) | numpy.isinf(reduced_data)).any()


class DiscreteStatePriorEstimator(_DiscreteParameterEstimator):
    """Estimate discrete state priors from observation sequences in Baum-Welch
    training, modeling these as an :class:`~minihmm.factors.ArrayFactor`
    """

    def reduce_data(self, my_obs, obs_logprob, forward, backward, scale_factors, ksi):
        """Collect data from a single observation sequence and reduce it to a
        form amenable for factor construction

        Parameters
        ----------
        my_obs : list-like
            Observation sequence

        obs_logprob : float
            Observation logprob, from
            :meth:`~minihmm.hmm.FirstOrderHMM.forward_backward`

        forward : numpy.ndarray
            Scaled forward probability matrix, from
            :meth:`~minihmm.hmm.FirstOrderHMM.forward_backward`

        backward : numpy.ndarray
            Scaled backward probability matrix, from
            :meth:`~minihmm.hmm.FirstOrderHMM.forward_backward`

        scale_factors : numpy.ndarray
            Scale factors used in scaling `forward` and `backward` to
            floating-point friendly sizes, from
            :meth:`~minihmm.hmm.FirstOrderHMM.forward_backward`

        ksi : numpy.ndarray
            `MxNxT` array describing the full probability of being in state `M`
            at time `t` and state `N` at time `t+1`. From
            :meth:`~minihmm.hmm.FirstOrderHMM.forward_backward`


        Returns
        -------
        numpy.ndarray
            State prior matrix for given observation sequence
        """
        return ksi[0, :, :].sum(1)

    def construct_factors(self, model, reduced_data, noise_weight=0, pseudocount_weight=1e-10):
        """Construct discrete transition factor for an HMM using reduced data
        from observation sequences

        model : :class:`~minihmm.hmm.FirstOrderHMM` or subclass

        reduced_data : numpy.ndarray
            sufficient statistics for observations, from
            :meth:`DiscreteStatePriorEstimator.reduce_data`

        noise_weight : float, optional
            weight of noise to add, relative to number of of observations (e.g.
            transition counts, state prior counts, emission counts, et c) in
            data set. (Default: 0)

        pseudocount_weight : float, optional
            weight of pseudocounts to add, relative to number of of
            observations (transition counts, state prior counts, emission
            counts, et c) in data set (Default: 1e-8)

        Returns
        -------
        :class:`~minihmm.factors.ArrayFactor`
            State prior probability factor
        """
        pi = sum(reduced_data)
        pi_sum = pi.sum()
        pi += get_model_noise(pi, noise_weight)
        pi += (pseudocount_weight * pi_sum / len(pi))
        pi_normed = pi / pi.sum()
        state_priors = ArrayFactor(pi_normed)
        return state_priors


class DiscreteTransitionEstimator(_DiscreteParameterEstimator):
    """Estimate transitions between states from observation sequences in Baum-Welch
    training. Constructs a MatrixFactor
    """

    def reduce_data(self, my_obs, obs_logprob, forward, backward, scale_factors, ksi):
        """Collect data from a single observation sequence and reduce it to a form
        amenable for factor construction

        Parameters
        ----------
        my_obs : list-like
            Observation sequence

        obs_logprob : float
            Observation logprob, from
            :meth:`minihmm.hmm.FirstOrderHMM.forward_backward`

        forward : numpy.ndarray
            Scaled forward probability matrix, from
            :meth:`~minihmm.hmm.FirstOrderHMM.forward_backward`

        backward : numpy.ndarray
            Scaled backward probability matrix, from
            :meth:`~minihmm.hmm.FirstOrderHMM.forward_backward`

        scale_factors : numpy.ndarray
            Scale factors used in scaling `forward` and `backward` to
            floating-point friendly sizes, from
            :meth:`~minihmm.hmm.FirstOrderHMM.forward_backward`

        ksi : numpy.ndarray
            `MxNxT` array describing the full probability of being in state `M`
            at time `t` and state `N` at time `t+1`. From
            :meth:`~minihmm.hmm.FirstOrderHMM.forward_backward`


        Returns
        -------
        numpy.ndarray
            Transition matrix for given observation sequence
        """
        return ksi.sum(0)

    def construct_factors(self, model, reduced_data, noise_weight=0, pseudocount_weight=1e-10):
        """Construct discrete transition factor for an HMM using reduced data
        from observation sequences

        model : :class:`~minihmm.hmm.FirstOrderHMM` or subclass

        reduced_data : numpy.ndarray
            sufficient statistics for observations, from
            :meth:`DiscreteTransitionEstimator.reduce_data`

        noise_weight : float, optional
            weight of noise to add, relative to number of of observations (e.g.
            transition counts, state prior counts, emission counts, et c) in
            data set. (Default: 0)

        pseudocount_weight : float, optional
            weight of pseudocounts to add, relative to number of of
            observations (transition counts, state prior counts, emission
            counts, et c) in data set (Default: 1e-8)

        Returns
        -------
        :class:`~minihmm.factors.MatrixFactor`
            Transition probability factor
        """
        A = sum(reduced_data)
        A_sum = A.sum()
        A += get_model_noise(A, noise_weight)
        A += (pseudocount_weight * A_sum / len(A.ravel()))
        A_normed = (A.T / A.sum(1)).T
        return MatrixFactor(A_normed)


class DiscreteEmissionEstimator(_DiscreteParameterEstimator):
    """Estimate discrete emissions from observation sequences in Baum-Welch
    training, modeling these as a series of ArrayFactos
    """

    def __init__(self, num_symbols):
        """Create a DiscreteEmissionEstimator

        Parameters
        ----------
        num_symbols : int
            Number of symbols that can be emitted
        """
        self.num_symbols = num_symbols
        _DiscreteParameterEstimator.__init__(self)

    def reduce_data(self, my_obs, obs_logprob, forward, backward, scale_factors, ksi):
        """Collect data from a single observation sequence and reduce it to a
        form amenable for factor construction

        Parameters
        ----------
        my_obs : list-like
            Observation sequence

        obs_logprob : float
            Observation logprob, from :meth:`~minihmm.hmm.FirstOrderHMM.forward_backward`

        forward : numpy.ndarray
            Scaled forward probability matrix, from
            :meth:`~minihmm.hmm.FirstOrderHMM.forward_backward`

        backward : numpy.ndarray
            Scaled backward probability matrix, from
            :meth:`~minihmm.hmm.FirstOrderHMM.forward_backward`

        scale_factors : numpy.ndarray
            Scale factors used in scaling `forward` and `backward` to
            floating-point friendly sizes, from
            :meth:`~minihmm.hmm.FirstOrderHMM.forward_backward`

        ksi : numpy.ndarray
            `MxNxT` array describing the full probability of being in state `M`
            at time `t` and state `N` at time `t+1`. From
            :meth:`~minihmm.hmm.FirstOrderHMM.forward_backward`

        Returns
        -------
        numpy.ndarray
            Emission matrix for given observation sequence
        """
        num_states = ksi.shape[1]  # ksi is time x states x states

        my_E = numpy.zeros((num_states, self.num_symbols), dtype=float)
        postprob = forward * backward
        tmp = postprob * scale_factors[:, None]
        for i in range(num_states):
            for k in range(self.num_symbols):
                my_E[i, k] += tmp[my_obs == k, i].sum()

        return my_E

    def construct_factors(self, model, reduced_data, noise_weight=0, pseudocount_weight=1e-10):
        """Construct discrete emission factor for an HMM using reduced data from
        observation sequences

        model : :class:`~minihmm.hmm.FirstOrderHMM` or subclass

        reduced_data : numpy.ndarray
            sufficient statistics for observations, from
            :meth:`DiscreteEmissionEstimator.reduce_data`

        noise_weight : float, optional
            weight of noise to add, relative to number of of observations (e.g.
            transition counts, state prior counts, emission counts, et c) in
            data set. (Default: 0)

        pseudocount_weight : float, optional
            weight of pseudocounts to add, relative to number of of
            observations (transition counts, state prior counts, emission
            counts, et c) in data set (Default: 1e-8)

        Returns
        -------
        list
            list of :class:`~minihmm.factors.ArrayFactor` objects, representing
            emission probabilities for each state
        """
        E = sum(reduced_data)
        E_sum = E.sum()
        E += get_model_noise(E, noise_weight)
        E += (pseudocount_weight * E_sum / len(E.ravel()))
        E_normed = (E.T / E.sum(1)).T
        emission_factors = []
        for i in range(E_normed.shape[0]):
            emission_factors.append(ArrayFactor(E_normed[i, :]))

        return emission_factors


class PseudocountStatePriorEstimator(DiscreteStatePriorEstimator):
    """Abstract class. Subclass and define `cls.pseudocount_array` to estimate
    state priors using arbitrarily distributed pseudocounts in Baum-Welch
    training
    """

    def __init__(self, pseudocount_array):
        """
        model : :class:`~minihmm.hmm.FirstOrderHMM`
            HMM to which this estimator will be attached

        pseudocount_array : float or numpy.ndarray
            Scalar (if evenly applying pseudocounts) or numpy array (if
            assymetrically weighting pseudocounts) pseudocounts to apply during
            estimation.
        """
        self.pseudocount_array = copy.deepcopy(pseudocount_array)
        DiscreteStatePriorEstimator.__init__(self)

    def construct_factors(self, model, reduced_data, noise_weight=0, pseudocount_weight=1e-10):
        """Construct discrete state prior factor for an HMM using reduced data
        from observation sequences

        Parameters
        ----------
        model : :class:`~minihmm.hmm.FirstOrderHMM` or subclass

        reduced_data : numpy.ndarray
            sufficient statistics for observations, from
            :meth:`PseudocountStatePriorEstimator.reduce_data`

        noise_weight : float, optional
            weight of noise to add, relative to number of of observations (e.g.
            transition counts, state prior counts, emission counts, et c) in
            data set. (Default: 0)

        pseudocount_weight : float, optional
            weight of pseudocounts to add, relative to number of of observations
            (transition counts, state prior counts, emission counts, et c)
            in data set (Default: 1e-8)

        Returns
        -------
        :class:`~minihmm.factors.MatrixFactor`
            Transition probability factor
        """
        pi = sum(reduced_data)
        pi_sum = pi.sum()
        pi += get_model_noise(pi, noise_weight, assymetric_weights=self.pseudocount_array)
        pi += (pseudocount_weight * pi_sum * self.pseudocount_array) / self.pseudocount_array.sum()
        pi_normed = pi / pi.sum()
        state_priors = ArrayFactor(pi_normed)
        return state_priors


class PseudocountTransitionEstimator(DiscreteTransitionEstimator):
    """Abstract class. Subclass and define `cls.pseudocount_array` to estimate
    state transitions using arbitrarily distributed pseudocounts in Baum-Welch
    training
    """

    def __init__(self, pseudocount_array):
        """
        Parameters
        ----------
        pseudocount_array : float or numpy.ndarray
            Scalar (if evenly applying pseudocounts) or numpy array (if
            assymetrically weighting pseudocounts) pseudocounts to apply during
            estimation.
        """
        self.pseudocount_array = copy.deepcopy(pseudocount_array)
        DiscreteTransitionEstimator.__init__(self)

    def construct_factors(self, model, reduced_data, noise_weight=0, pseudocount_weight=1e-10):
        """Construct discrete transition factor for an HMM using reduced data
        from observation sequences

        model : :class:`~minihmm.hmm.FirstOrderHMM` or subclass

        reduced_data : numpy.ndarray
            sufficient statistics for observations, from
            :meth:`PseudocountTransitionEstimator.reduce_data`

        noise_weight : float, optional
            weight of noise to add, relative to number of of observations (e.g.
            transition counts, state prior counts, emission counts, et c) in
            data set. (Default: 0)

        pseudocount_weight : float, optional
            weight of pseudocounts to add, relative to number of of
            observations (transition counts, state prior counts, emission
            counts, et c) in data set (Default: 1e-8)

        Returns
        -------
        :class:`~minihmm.factors.MatrixFactor`
            Transition probability factor
        """
        A = sum(reduced_data)
        A_sum = A.sum()
        A += get_model_noise(A, noise_weight, assymetric_weights=self.pseudocount_array)
        A += (pseudocount_weight * A_sum * self.pseudocount_array) / self.pseudocount_array.sum()
        A_normed = (A.T / A.sum(1)).T
        return MatrixFactor(A_normed)


class PseudocountEmissionEstimator(DiscreteEmissionEstimator):
    """Abstract class. Subclass and define `cls.pseudocount_array` to estimate
    emissions using arbitrarily distributed pseudocounts in Baum-Welch training
    """

    def __init__(self, model, num_symbols, pseudocount_array):
        """
        model : :class:`~minihmm.hmm.FirstOrderHMM`
            HMM to which this estimator will be attached

        pseudocount_array : float or numpy.ndarray
            Scalar (if evenly applying pseudocounts) or numpy array (if
            assymetrically weighting pseudocounts) pseudocounts to apply during
            estimation.
        """
        self.pseudocount_array = copy.deepcopy(pseudocount_array)
        DiscreteEmissionEstimator.__init__(self, model, num_symbols)

    def construct_factors(self, model, reduced_data, noise_weight=0, pseudocount_weight=1e-10):
        """Construct discrete emission factor for an HMM using reduced data from
        observation sequences

        model : :class:`~minihmm.hmm.FirstOrderHMM` or subclass

        reduced_data : numpy.ndarray
            sufficient statistics for observations, from
            :meth:`PseudocountEmissionEstimator.reduce_data`

        noise_weight : float, optional
            weight of noise to add, relative to number of of observations (e.g.
            transition counts, state prior counts, emission counts, et c) in
            data set. (Default: 0)

        pseudocount_weight : float, optional
            weight of pseudocounts to add, relative to number of of
            observations (transition counts, state prior counts, emission
            counts, et c) in data set (Default: 1e-8)

        Returns
        -------
        list
            list of :class:`~minihmm.factors.ArrayFactor` objects representing
            emission probabilities for each state
        """
        E = sum(reduced_data)
        E_sum = E.sum()
        E += get_model_noise(E, noise_weight, assymetric_weights=self.pseudocount_array)
        E += (pseudocount_weight * E_sum * self.pseudocount_array) / self.pseudocount_array.sum()
        E_normed = (E.T / E.sum(1)).T
        emission_factors = []
        for i in range(E_normed.shape[0]):
            emission_factors.append(ArrayFactor(E_normed[i, :]))

        return emission_factors


class TiedStatePriorEstimator(PseudocountStatePriorEstimator):
    """Estimate state prior probabilities, but tying (pooling data for and then
    jointly estimating) comparable states (e.g. all single states, all
    compound states).
    """

    def __init__(self, pseudocount_array, index_map):
        """
        pseudocount_array : float or numpy.ndarray
            Scalar (if evenly applying pseudocounts) or numpy array (if
            assymetrically weighting pseudocounts) pseudocounts to apply during
            estimation.

        index_map : numpy.ndarray
            a NUM_STATES-vector in which cells containing identical
            integer values have tied parameters.
        """
        self.index_map = copy.deepcopy(index_map)
        self.index_weights = numpy.array([(self.index_map == X).sum() \
                                          for X in range(0, 1+self.index_map.max())]
                                         )
        self.pseudocount_mask = (pseudocount_array > 0)
        PseudocountStatePriorEstimator.__init__(self, pseudocount_array)

    def construct_factors(self, model, reduced_data, noise_weight=0, pseudocount_weight=1e-10):
        """Construct state prior factor for an HMM using reduced data from
        observation sequences

        Parameters
        ----------
        model : :class:`~minihmm.hmm.FirstOrderHMM` or subclass

        reduced_data : numpy.ndarray
            sufficient statistics for observations, from
            :meth:`TiedStatePriorEstimator.reduce_data`

        noise_weight : float, optional
            weight of noise to add, relative to number of of observations (e.g.
            transition counts, state prior counts, emission counts, et c) in
            data set. (Default: 0)

        pseudocount_weight : float, optional
            weight of pseudocounts to add, relative to number of of
            observations (transition counts, state prior counts, emission
            counts, et c) in data set (Default: 1e-8)


        Returns
        -------
        :class:`~minihmm.factors.ArrayFactor`
           Tied state priors
        """
        pi_raw = sum(reduced_data)
        pi_sum = pi_raw.sum()
        reduced_vector = numpy.zeros(1 + max(self.index_map))

        for i, val in enumerate(pi_raw):
            reduced_vector[self.index_map[i]] += pi_raw[i]

        # add noise
        reduced_vector += get_model_noise(
            reduced_vector, noise_weight, assymetric_weights=self.index_weights
        )  # FIXME: THIS WILL ADD NOISE TO FORBIDDEN CELLS

        # divide each starting cell by number of destination cells
        reduced_vector /= self.index_weights

        # populate destination vector
        pi_proc = numpy.zeros_like(pi_raw, dtype=float)
        for i in range(len(pi_proc)):
            pi_proc[i] = reduced_vector[self.index_map[i]]

        # add pseudocounts
        pi_proc += (pseudocount_weight * pi_sum *
                    self.pseudocount_array) / self.pseudocount_array.sum()
        pi_proc *= self.pseudocount_mask  # re-zero forbidden cells that received noise during tying

        # normalize
        pi_proc /= pi_proc.sum()

        return ArrayFactor(pi_proc)


class TiedTransitionEstimator(PseudocountTransitionEstimator):
    """Estimate state prior probabilities, but tying (pooling data for and then
    jointly estimating) comparable states (e.g. all single states, all
    compound states).
    """

    def __init__(self, pseudocount_array, index_map):
        """
        pseudocount_array : float or numpy.ndarray
            Scalar (if evenly applying pseudocounts) or numpy array (if
            assymetrically weighting pseudocounts) pseudocounts to apply during
            estimation.

        index_map : numpy.ndarray
            a `NUM_STATES x NUM_STATES` table vector in which cells
            containing identical integer values have tied parameters.
        """
        self.index_map = copy.deepcopy(index_map)
        self.index_weights = numpy.array([(self.index_map == X).sum() \
                                          for X in range(0, 1 + self.index_map.max())]
                                         )
        self.pseudocount_mask = (pseudocount_array > 0)
        PseudocountTransitionEstimator.__init__(self, pseudocount_array)

    def construct_factors(self, model, reduced_data, noise_weight=0, pseudocount_weight=1e-10):
        """Construct transition factor for an HMM using reduced data from
        observation sequences

        model : :class:`~minihmm.hmm.FirstOrderHMM` or subclass

        reduced_data : numpy.ndarray
            sufficient statistics for observations, from
            :meth:`TiedTransitionEstimator.reduce_data`

        noise_weight : float, optional
            weight of noise to add, relative to number of of observations (e.g.
            transition counts, state prior counts, emission counts, et c) in
            data set. (Default: 0)

        pseudocount_weight : float, optional
            weight of pseudocounts to add, relative to number of of
            observations (transition counts, state prior counts, emission
            counts, et c) in data set (Default: 1e-8)

        Returns
        -------
        :class:`~minihmm.factors.MatrixFactor`
            Tied state transition probability table
        """
        A_raw = sum(reduced_data)
        A_sum = A_raw.sum()
        reduced_vector = numpy.zeros(1 + self.index_map.max())

        for i in range(A_raw.shape[0]):
            for j in range(A_raw.shape[1]):
                reduced_vector[self.index_map[i, j]] += A_raw[i, j]

        # add noise
        reduced_vector += get_model_noise(
            reduced_vector, noise_weight, assymetric_weights=self.index_weights
        )  # FIXME: THIS WILL ADD NOISE TO FORBIDDEN CELLS

        # divide each starting cell by number of destination cells
        reduced_vector /= self.index_weights

        # populate destination vector
        A_proc = numpy.zeros_like(A_raw, dtype=float)
        for i in range(A_proc.shape[0]):
            for j in range(A_proc.shape[1]):
                A_proc[i, j] = reduced_vector[self.index_map[i, j]]

        # add pseudocounts
        A_proc += (pseudocount_weight * A_sum *
                   self.pseudocount_array) / self.pseudocount_array.sum()
        A_proc *= self.pseudocount_mask  # re-zero forbidden cells that became zero via noise addition

        # normalize
        A_proc = (A_proc.T / A_proc.sum(1)).T

        return MatrixFactor(A_proc)


# TODO: implement model noise? by adding Gaussian noise around the observation vectors?
#        what would pseudocounts/regularization mean, if anything, in this context?
class UnivariateGaussianEmissionEstimator(AbstractProbabilityEstimator):
    """Estimate univariate Gaussian emissions from observation sequences in
    Baum-Welch training.
    """

    def is_valid(self, reduced_data):
        return not (numpy.isnan(reduced_data) | numpy.isinf(reduced_data)).any()

    def reduce_data(self, my_obs, obs_logprob, forward, backward, scale_factors, ksi):
        """Collect data from a single observation sequence and reduce it to a
        form amenable for factor construction

        Parameters
        ----------
        my_obs : list-like
            Observation sequence

        obs_logprob : float
            Observation logprob, from
            :meth:`~minihmm.hmm.FirstOrderHMM.forward_backward`

        forward : numpy.ndarray
            Scaled forward probability matrix, from
            :meth:`~minihmm.hmm.FirstOrderHMM.forward_backward`

        backward : numpy.ndarray
            Scaled backward probability matrix, from
            :meth:`~minihmm.hmm.FirstOrderHMM.forward_backward`

        scale_factors : numpy.ndarray
            Scale factors used in scaling `forward` and `backward` to
            floating-point friendly sizes, from
            :meth:`~minihmm.hmm.FirstOrderHMM.forward_backward`

        ksi : numpy.ndarray
            `MxNxT` array describing the full probability of being in state `M`
            at time `t` and state `N` at time `t+1`. From

        Returns
        -------
        numpy.ndarray
            Emission matrix for given observation sequence, in which each row
            is a model state. The first column is the estimated mean, the
            second the estimated variance, and the third, the number of points
            counted in each sequence
        """
        num_states = ksi.shape[1]
        my_E = numpy.zeros((num_states, 3))
        postprob = forward * backward
        tmp = postprob * scale_factors[:, None]
        for i in range(num_states):
            # n counted in mean
            num_observations = tmp[:, i].sum()
            my_E[i, 2] = num_observations

            # probability-weighted mean
            my_E[i, 0] = (tmp[:, i] * my_obs).sum() / num_observations

            # probability-weighted unbiased variance
            my_E[i, 1] = (tmp[:, i] * (my_obs - my_E[i, 0])**2).sum() / (num_observations - 1)

        return my_E

    def construct_factors(self, model, reduced_data, **ignored):
        """Construct discrete emission factor for an HMM using reduced data from
        observation sequences.

        Pooled variance estimation formula drawn from
        en.wikipedia.org/wiki/Pooled_variance

        Parameters
        ----------
        model : :class:`~minihmm.hmm.FirstOrderHMM` or subclass

        reduced_data : numpy.ndarray
            sufficient statistics for observations, from
            :meth:`UnivariateTEmissionEstimator.reduce_data`


        Returns
        -------
        list
            List of :class:`~minihmm.factors.ScipyDistributionFactor` objects
            representing Gaussian emission factors for each state
        """
        E = numpy.zeros_like(reduced_data[0])
        var_pool_numerators = numpy.zeros(E.shape[0])
        for my_E in reduced_data:
            E[:, 0] += my_E[:, 0] * my_E[:, 2]
            E[:, 2] += my_E[:, 2]
            var_pool_numerators += my_E[:, 1] * (my_E[:, 2] - 1)

        E[:, 0] /= E[:, 2]
        E[:, 1] = var_pool_numerators / (E[:, 2] - len(reduced_data))  #dof-1 var correction

        emission_factors = []
        for i in range(E.shape[0]):
            emission_factors.append(
                ScipyDistributionFactor(scipy.stats.norm, loc=E[i, 0], scale=E[i, 1]**0.5)
            )

        return emission_factors


# NOT TESTED
class UnivariateTEmissionEstimator(UnivariateGaussianEmissionEstimator):
    """Estimate univariate T-distributed emissions from observation sequences
    in Baum-Welch training.
    """

    def construct_factors(self, model, reduced_data, **ignored):
        """Construct discrete emission factor for an HMM using reduced data from
        observation sequences.

        Pooled variance estimation formula drawn from
        en.wikipedia.org/wiki/Pooled_variance

        Parameters
        ----------
        model : :class:`~minihmm.hmm.FirstOrderHMM` or subclass

        reduced_data : numpy.ndarray
            sufficient statistics for observations, from
            :meth:`UnivariateTEmissionEstimator.reduce_data`


        Returns
        -------
        list
            list of :class:`~minihmm.factors.ScipyDistributionFactor` objects
            representing emission factors for each state
        """
        E = numpy.zeros_like(reduced_data[0])
        n = numpy.zeros(E.shape[0])
        var_pool_numerators = numpy.zeros(E.shape[0])
        for my_E in reduced_data:
            E[:, 0] += my_E[:, 0] * my_E[:, 2]
            E[:, 2] += my_E[:, 2]
            var_pool_numerators += my_E[:, 1] * (my_E[:, 2] - 1)
            n += my_E[:, 2]

        E[:, 0] /= n
        E[:, 1] = var_pool_numerators / (n - len(reduced_data))

        emission_factors = []
        for i in range(E.shape[0]):
            emission_factors.append(
                ScipyDistributionFactor(
                    scipy.stats.t,
                    n[i] - 1,  # dof
                    loc=E[i, 0],
                    scale=E[i, 1]**0.5
                )
            )

        return emission_factors



#!/usr/bin/env python
"""Utilities for changing representations of observation sequences and/or models.
The primary functionality of the tools in this module is to translate Nth-order
HMMs into equivalent first order HMMs.


Notes on model reduction
------------------------
In a high-order model, the ends of sequences introduce inhomogeneities, similar
to state priors in first-order models. For example, in a 3rd order model, the
first two states must be handled separately from those remaining:

.. math::

   P = p(S_1) * p(S_2 | S_1) * p(S_3 | S_2, S_1) * p(S_4 | S_3, S_2) *  ... * p(S_N |, S_{N-2}, S_{N-1})

Rather than adding extra probability tables to represent these inhomogeneities,
:mod:`minihmm` creates dummy states. For a third-order model, we would need
to add two dummy start states in high-order space, $ S_{dummy1} $ and $ S_{dummy2} $.

We rephrase the model above as:

.. math::

   P = p(S_1 | S_{dummy2}, S_{dummy1}) * p(S_2 | S_1, S_{dummy2}, ) * p(S_3 | S_2, S_1) * p(S_4 | S_3, S_2) *  ... * p(S_N |, S_{N-2}, S_{N-1})

Then:

 - The prior probabilities of every state pair are set to zero, except
   $ p(S_{dummy2}, S_{dummy1}) $, which is  set to one.

 - The transition probabilities of p(S_i | S_{dummy2}, S_{dummy1}) are set to
   the prior probabilities of P(S_i) in the previous model description

 - Transition probabilities of p(S_i | S_{i-1}, S_{dummy2}) are set to the
   observed values for P(S_i | S_{i-1}) in the previous model description

:mod:`minihmm` contains tools for creating these dummy states, and remapping
state sequences accordingly.



Workflow
--------

A number of different workflows are possible depending upon what type of
starting information is available:


If state path for each observation sequence are known
.....................................................

In this case, it is easiest to transcode the state sequences into first-order
space, and then calculate the model parameters from the known state sequences.
To do this, call :meth:`ModelReducer.lower_stateseq_orders`, to re-map state
sequences into the reduced model space, then calculate transition probabilities
as for a first order model.


If series of states are unknown, and training is required
.........................................................

In this case, instantiate a :class:`~minihmm.represent.ModelReducer` and either
(1a) create a naive model using
:meth:`~minihmm.represent.ModelReducer.get_random_model`, or, better, (1b) remap
parameters from a related first-order model using
:meth:`~minihmm.represent.ModelReducer.remap_from_first_order` and then (2)
train the resulting HMM using standard method (e.g.
:func:`~minihmm.training.train_baum_welch`), tying emission factors to improve
fitting (see :meth:`~minihmm.represent.ModelReducer.get_emission_mapping`.



"""
import warnings
import itertools
import copy
import numpy

import jsonpickle
import jsonpickle.ext.numpy
jsonpickle.ext.numpy.register_handlers()

from minihmm.hmm import FirstOrderHMM
from minihmm.factors import ArrayFactor, MatrixFactor

from scipy.sparse import (lil_matrix, dok_matrix, coo_matrix)

#===============================================================================
# Model translation
#===============================================================================


def _get_modelreducer_from_dict(dtmp):
    """Revive a :class:`ModelReducer` from a dictionary made by
    :meth:`~ModelReducer._to_dict`

    Parameters
    ----------
    dtmp : dict
        Dictionary exported by :meth:`~ModelReducer._to_dict`
        
    Returns
    -------
    ModelReducer
        Revived model
    """
    # code note: logic has to be in an independent function as opposed
    # to a static method in order to enable its use in __reduce__
    return ModelReducer(dtmp["starting_order"], dtmp["num_states"], hmm=dtmp.get("hmm", None))


class ModelReducer(object):
    """Utility class for reducing high-order HMMs to equivalent first-order HMMs.


    Attributes
    ----------
    starting_order : int
        Order of starting model

    high_order_states : int
        Number of states, in high-order space

    low_order_states : int
        Number of states, in fist-order space

    high_states_to_low : dict
        Dictionary mapping high-order states to tuples of low-order states

    low_states_to_high : dict
        Dicitonary mapping tuples of low-order states to high-order states

    hmm : :class:`~minihmm.hmm.FirstOrderHMM`
        Associated first-order HMM. At the moment, this must be constructed by
        the user.. Will be used for sampling, decoding, et c
    """

    def __init__(self, starting_order, num_states, hmm=None):

        if num_states < 1:
            raise ValueError("Must have >= 1 state in model.")
        if starting_order < 1:
            raise ValueError("Cannot reduce order of 0-order model")
        elif starting_order == 1:
            warnings.warn("Reducing a first-order model doesn't make much sense.", UserWarning)

        self.starting_order = starting_order
        self.high_order_states = num_states
        self._dummy_states = self._get_dummy_states()
        self.high_states_to_low, self.low_states_to_high = self._get_state_mapping()
        self.low_order_states = len(self.low_states_to_high)
        self._hmm = hmm

    @property
    def hmm(self):
        if self._hmm is None:
            cname = self.__class__.__name__
            raise ValueError(
                "No HMM associated with %s. Please set `this.hmm` to something." % cname
            )

        return self._hmm

    @hmm.setter
    def hmm(self, value):
        if not isinstance(value, FirstOrderHMM) and value is not None:
            raise ValueError(
                "`hmm` must be a valid FirstOrderHMM. Instead got type '%s'" %
                (type(value).__name__)
            )

        self._hmm = value

    def __str__(self):
        return "<%s order=%s high_states=%s hmm=%s>" % (
            self.__class__.__name__, self.starting_order, self.high_order_states,
            self._hmm is not None
        )

    def __eq__(self, other):
        """Test whether `self` is equal to `other`, defined as equality of:

         - number of high order states
         - starting model order
         - if an HMM is defined for either `self` or `other`, both must be defined,
           and have equal parameters

        Returns
        -------
        bool
            `True` if `self` equals `other`, otherwise `False`
        """
        if self.starting_order != other.starting_order \
           or self.high_order_states != self.high_order_states \
           or self.low_order_states != other.low_order_states:
            return False

        if self._hmm is None:
            if other._hmm is None:
                return True
        else:
            if other._hmm is None:
                return False
            return self.hmm == other.hmm

    def __repr__(self):
        return str(self)

    def __reduce__(self):
        """Define pickling and unpickling methods for `self`"""
        return _get_modelreducer_from_dict, (self._to_dict(), )

    def _to_dict(self):
        """Convenience method to export minimal elements required for pickling

        Returns
        -------
        dict
            Dictionary representation of `self`
        """
        dtmp = {
                "starting_order" : self.starting_order,
                "num_states"     : self.high_order_states,
        } # yapf: disable
        if self._hmm is not None:
            dtmp["hmm"] = self._hmm

        return dtmp

    @staticmethod
    def _from_dict(dtmp):
        """Revive a :class:`ModelReducer` from a dictionary made by
        :meth:`ModelReducer._to_dict`

        Returns
        -------
        ModelReducer
            Revived model
        """
        # code note: logic has to be in an independent function as opposed
        # to a static method in order to enable its use in __reduce__
        return _get_modelreducer_from_dict(dtmp)

    def to_json(self):
        """Convert `self` to a JSON blob

        Returns
        -------
        str
            JSON blob of `self`
        """
        return jsonpickle.encode(self)

    @staticmethod
    def from_json(stmp):
        """Revive a :class:`ModelReducer` from a string-format JSON blob"""
        return jsonpickle.decode(stmp)

    @staticmethod
    def transcode_sequence(sequence, alphadict):
        """Transcode a single sequence from one alphabet to another

        Parameters
        ----------
        sequence : list-like
            A sequence

        alphadict : dict-like
            Dictionary mapping symbols in input sequence to symbols in output
            sequence

        Returns
        -------
        :class:`numpy.ndarray`
            Transcoded sequence
        """
        return numpy.array([alphadict[X] for X in sequence])

    @staticmethod
    def transcode_sequences(sequences, alphadict):
        """Transcode a set of sequences from one alphabet to another

        Parameters
        ----------
        sequences : iterable of list-like
            Iterable of sequences (each, itself, an iterable like a list et c)

        alphadict : dict-like
            Dictionary mapping symbols in input sequence to symbols in output
            sequence

        Returns
        -------
        list of :class:`numpy.ndarray`
            List of transcoded sequences
        """
        return [ModelReducer.transcode_sequence(X, alphadict) for X in sequences]

    def _get_dummy_states(self):
        """Create sorted lists of dummy states required to reduce `self` to
        an equivalent first-order model.

        By convention, dummy states are given negative indices in high-order
        space.  Don't rely on this- it may change in the future

        Parameters
        ----------
        starting_order : int
            Starting order of HMM/MM

        Returns
        -------
        list
            List of new dummy states, sorted
        """
        return list(range(1 - self.starting_order, 0))

    def _get_state_mapping(self):
        """Create dicts mapping states between a high-order and a first-order
        HMM.

        Returns
        -------
        :class:`dict`
            Forward state map, mapping high-order states to tuples of
            equivalent low-order states

        :class:`dict`
            Reverse state map, mapping new first-order states to tuples of
            high-order states
        """
        forward = {}
        reverse = {}
        states = list(range(self.high_order_states))

        c = 0

        # create maps for newly added start and end states
        for idx in range(self.starting_order - 1):
            root = self._dummy_states[idx:]
            for symbol in itertools.product(*([states] * (idx + 1))):
                tup = tuple(root + list(symbol))
                forward[tup] = c
                reverse[c] = tup
                c += 1

        # combinations of true states
        for n, symbol in enumerate(itertools.product(*([states] * self.starting_order))):
            forward[symbol] = n + c
            reverse[c + n] = symbol

        return forward, reverse

    def _get_stateseq_tuples(self, state_seqs):
        """Remap a high-order sequence of states into tuples for use in a
        low-order model, adding dummy start states

        Notes
        -----
        This does *not* remap those tuples into lower state spaces. Use
        :func:`lower_stateseq_orders`, which wraps this function, for that


        Parameters
        ----------
        state_seqs : list of list-like
            List of state sequences

        Returns
        -------
        list of lists
            List of tuples of state sequences
        """
        dummy_states = self._dummy_states
        starting_order = self.starting_order
        outseqs = []
        for n, inseq in enumerate(state_seqs):
            if (numpy.array(inseq) < 0).any():
                raise ValueError("Found negative state label in input sequence %s!" % n)

            baseseq = dummy_states + list(inseq)
            outseqs.append(
                [
                    tuple(baseseq[idx:idx + starting_order])
                    for idx in range(0,
                                     len(baseseq) - starting_order + 1)
                ]
            )

        return outseqs

    def lower_stateseq_orders(self, state_seqs):
        """Map a high-order sequence of states into an equivalent first-order
        state sequence, creating dummy states as necessary and dictionaries
        that map states between high and first-order spaces.


        Parameters
        ----------
        state_seqs : list of list-like
            List of state sequences, each given in high-order space

        Returns
        -------
        list of :class`numpy.ndarray`
            List of state sequences, represented in first-order space


        See also
        --------
        raise_stateseq_orders
        """
        dummy_states = self._dummy_states
        state_map = self.high_states_to_low

        tuple_seqs = self._get_stateseq_tuples(state_seqs)
        return ModelReducer.transcode_sequences(tuple_seqs, state_map)

    def raise_stateseq_orders(self, state_seqs):
        """Map a state sequence from first-order space back to original
        high-order space

        Parameters
        ----------
        state_seqs : list
            List of high-order state sequences

        Returns
        -------
        list of :class:`numpy.ndarray`
            State sequences, in high-order space


        See also
        --------
        lower_stateseq_orders
        """
        ltmp = []
        for t in self.transcode_sequences(state_seqs, self.low_states_to_high):
            ltmp.append([X[-1] for X in t])

        return ltmp

    def viterbi(self, emissions):
        """Finds the most likely state sequence underlying a set of emissions
        using the Viterbi algorithm.

        See http://en.wikipedia.org/wiki/Viterbi_algorithm

        Parameters
        ----------
        emissions : numpy.ndarray
            Sequence of observations

        Returns
        -------
        `viterbi_states`
            :class:`numpy.ndarray`. Decoded labels for each position in
            `emissions[start:end]`
        """
        raw = self.hmm.viterbi(emissions)["viterbi_states"]
        high = self.raise_stateseq_orders([raw])[0]
        return high

    def posterior_decode(self, emissions):
        """Find the most probable state for each individual state in the
        sequence of emissions, using posterior decoding. Note, this objective
        is distinct from finding the most probable sequence of states for all
        emissions, as is given in Viterbi decoding. This alternative may be
        more appropriate when multiple paths have similar probabilities.


        Parameters
        ----------
        emissions : numpy.ndarray
            Sequence of observations


        Returns
        -------
        numpy.ndarray
            An array of dimension `[t x 1]` of the most likely states at each point `t`
        """
        raw, _ = self.hmm.posterior_decode(emissions)
        return self.raise_stateseq_orders([raw])[0]

    def sample(self, emissions, num_samples):
        """Sample state sequences from the distribution P(states | emissions),
        by tracing backward through the matrix of forward probabilities.
        See Durbin1997 ch 4.3, section "Probabilistic sampling of aligments"

        Parameters
        ----------
        emissions : numpy.ndarray
            Sequence of observations

        num_samples : int, optional
            Number of state paths to generate (Default: 1)


        Returns
        -------
        list
            List of state paths
        """
        raw_paths = self.hmm.sample(emissions, num_samples=num_samples)
        return self.raise_stateseq_orders(raw_paths)

    def generate(self, length):
        """Generates a random sequence of states and emissions from the HMM

        Parameters
        ----------
        length : int
            Length of sequence to generate


        Returns
        -------
        numpy.ndarray
            Array of dimension `[t x 1]` indicating the HMM state at each
            timestep

        numpy.ndarray
            Array of dimension `[t x Q]` indicating the observation at each
            timestep.  `Q = 1` for univariate HMMs, or more than 1 if
            observations are multivariate.

        float
            Joint log probability of generated state and observation sequence.
            **Note**: this is different from the log probability of the
            observation sequence alone, which would be the sum of its joint
            probabilities with all possible state sequences.


        Notes
        -----
        The HMM can only generate sequences if all of its EmissionFactors are
        generative. I.e. if using :class:`minihmm.factors.FunctionFactor` or
        :class:`minihmm.factors.LogFunctionFactor` , generator functions must
        be specified at their instantiation.
        """
        raw_path, obs, logprob = self.hmm.generate(length)
        high_path = self.raise_stateseq_orders([raw_path])[0]
        return high_path, obs, logprob

    def joint_path_logprob(self, path, emissions):
        """Return log P(path, emissions) evaluated under this model

        Parameters
        ----------
        path : list-like
            Sequence of states

        emissions : list-like
            Sequence of observations

        Returns
        -------
        float
            Log probability of P(path, emissions)
        """
        lower_path = self.lower_stateseq_orders([path])[0]
        return self.hmm.joint_path_logprob(lower_path, emissions)

    def remap_emission_factors(self, emission_probs):
        """Map emission probabilities from high-order space to equivalent
        reduced space

        Parameters
        ----------
        emission_probs : list of :class:`~minihmm.factors.AbstractFactor`
            List of emission probabilities, indexed by state in high-order
            space.

        Returns
        -------
        list
            list of Factors, indexed by states in equivalent first-order
        """
        # make empty list length of new states
        ltmp = [None] * len(self.low_states_to_high)

        # populate list
        for newstate, state_tuple in self.low_states_to_high.items():
            ltmp[newstate] = emission_probs[state_tuple[-1]]

        return ltmp

    # convert to csr, csc, or dense before computations depending on which needed
    # serialize coomat as coomat.row, coomat.col, coomat.data
    def get_pseudocount_arrays(self):
        """Return a valid pseudocount array for state priors and for transition
        tables in first-order space, where *valid* stipulates that cells
        corresponding transitions that cannot exist in high-order space are set
        to zero.

        Returns
        -------
        :class:`numpy.ndarray`
            Array of state prior pseudocounts

        :class:`scipy.sparse.coo_matrix`
            Sparse matrix of transition pseudocounts
        """
        state_priors = numpy.zeros(self.low_order_states)
        for i in range(self.high_order_states):
            state_priors[self.high_states_to_low[tuple(self._dummy_states + [i])]] = 1

        row_ords = []
        col_ords = []
        for lstate, hseq in self.low_states_to_high.items():
            stub = hseq[1:]
            for i in range(self.high_order_states):
                next_state = self.high_states_to_low[tuple(list(stub) + [i])]
                row_ords.append(lstate)
                col_ords.append(next_state)

        vals = [1] * len(row_ords)

        return state_priors, coo_matrix((vals, (row_ords, col_ords)))

    def get_emission_mapping(self):
        """Generate an array mapping emission factors in translated low-order
        models to their corresponding emission factors in native, high-order
        states. In training, the parameters for these groups of emissions
        should be tied, as they are equivalent.

        Returns
        -------
        :class:`numpy.ndarray`
            Array of length `len(self.high_states_to_low)`, indexed by state in
            low-order space, mapping low-order states to corresponding native
            high-order states

        """
        # can rely on lexicographic pairing of high-order state tuples
        # and translated, low-order state numbers to produce correct tying vector
        return numpy.array([X[-1] for X in sorted(self.high_states_to_low)])


#    # TODO
#    def get_random_model(self):
#        """
#        See also
#        --------
#        :meth:`ModelReducer.remap_from_first_order`
#        """
#        pass

    def remap_from_first_order(self, native_hmm):
        """Remap parameters from a native first order HMM onto a first-order
        translation of a high-order HMM, in order to, for example, provide a
        reasonable non-random starting point for refinement training of the
        high-order HMM.

        Parameters
        ----------
        native_hmm : :class:`minihmm.hmm.FirstOrderHMM`
            Native, first-order HMM, preferably with trained parameters

        Returns
        -------
        :class:`~minihmm.hmm.FirstOrderHMM`
            First-order representation of the high-order HMM structure
            described by `self`, with parameters from `native_hmm` remapped
            into corresponding positions.
        """
        htl = self.high_states_to_low

        # check that number of states is compatible
        if self.high_order_states != native_hmm.num_states:
            raise ValueError(
                "Native HMM (%d states), has different number of states than `self` (%d states)" %
                (native_hmm.num_states, self.high_order_states)
            )

        # For transitions
        # Each high-order state transitiono  `(n-i, ... , n-1) ->  (n-i+1 , ... , n)`
        # should be mapped to appropiate transformations of the parameters (n - 1 , n)

        # For state priors and emission probabilities
        # each high-order state (n-i, ...,  n) should be given the parameters matching
        # native state `n`

        # will need to make appropriate state-tying matrices for emissions, as well
        sp_source = native_hmm.state_priors.data
        sp_dest = numpy.zeros(self.low_order_states, dtype=float)

        trans_source = native_hmm.trans_probs.data
        trans_dest = numpy.zeros((self.low_order_states, self.low_order_states), dtype=float)

        em_source = native_hmm.emission_probs
        em_dest = [None] * self.low_order_states

        for my_tuple, trans_state in htl.items():
            native_state = my_tuple[-1]
            sp_dest[trans_state] = sp_source[native_state]
            em_dest[trans_state] = copy.deepcopy(em_source[native_state])

            for next_native_state in range(self.high_order_states):
                next_tuple = tuple(list(my_tuple)[1:] + [next_native_state])
                next_trans_state = htl[next_tuple]
                trans_dest[trans_state, next_trans_state] = trans_source[native_state,
                                                                         next_native_state]

        # renormalize
        sp_dest /= sp_dest.sum()
        sp_dest = ArrayFactor(sp_dest)

        # shoudln't have to renormalize; check this
        trans_dest = (trans_dest.T / trans_dest.sum(1)).T
        trans_dest = MatrixFactor(trans_dest)

        return FirstOrderHMM(state_priors=sp_dest, emission_probs=em_dest, trans_probs=trans_dest)
